
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_starpar_lib.h"

/**@{*/

/*----------------------------------------------------------------------------*/
/**
 * @defgroup espda_starpar_lib
 *
 * TBD
 */
/*----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*/
/**
  @brief    Compile line ratios from a list of measured line equivalent widths
  @param    lares   array with wavelength of lines measured
  @param    ewares  array with EW of lines measured
  @param    na      length of measured EW lines
  @param    nf      length of ratios array lines
  @param    line1   array with line 1 of ratios
  @param    line2   array with line 2 of ratios
  @param    lratios array to contain the ratios values
  @return   no return 

  This function computes the ratios values that can be computed for the inputed
  measured EW lines.

*/
/*------------------------------------------------------------------------------*/
void esp_compile_ratios(double lares[], 
						double ewares[], 
						long na, 
						long nf, 
						double line1[], 
						double line2[], 
						double lratios[]){
	long i,li,lj,j;
	li=-1;
	lj=-1;

	for (i=0; i < nf; i++){	
		li=-1;
		lj=-1;
		for (j=0; j < na; j++)	{
			if (fabs(lares[j]-line1[i]) < 0.01)	li=j;
			if (fabs(lares[j]-line2[i]) < 0.01)	lj=j;
		}
		if (li == -1 || lj == -1) lratios[i]=-1; else lratios[i]=ewares[li]/ewares[lj];
	}
}



/*----------------------------------------------------------------------------*/
/**
  @brief    This function computes the value of the effective temperature based
			on the computed  line ratios.
  @param    lineratio_calib_table  cpl table with the calibrations for each ratio
  @param    line_ew_table	       cpl table with the EW measurements for the lines
  @param    teff_table      	   cpl table to store the computed teff
  @return   cpl error code of execution

  This function computes the teff for each ratio. Computes the average value 
  removing some outliers and return the resulting teff, its standard deviation,
  and the number of line ratios used in the computation. It also checks the number 
  of independent ratios (check Sousa et al. 2010A&A...512A..13S).

*/
/*------------------------------------------------------------------------------*/
cpl_error_code espda_calib_teff(cpl_table* lineratio_calib_table, 
							  cpl_table* line_ew_table, 
							  cpl_table** teff_table) {

	cpl_errorstate prestate = cpl_errorstate_get();

	cpl_size nl = cpl_table_get_nrow(lineratio_calib_table);
	int i;
	double stddev[nl];
	double line1[nl],line2[nl],nfit[nl];

	double calib[nl][4];

	char ele1[nl][6],ele2[nl][6];


	for (i=0;i<nl;i++) 	{
    	stddev[i] = cpl_table_get_double(lineratio_calib_table,"STDDEV_FIN",i,NULL);
		line1[i] = cpl_table_get_double(lineratio_calib_table,"WAVE1",i,NULL);
    	line2[i] = cpl_table_get_double(lineratio_calib_table,"WAVE2",i,NULL);
		strcpy (ele1[i], cpl_table_get_string(lineratio_calib_table,"ELEMENT1",i));
		strcpy (ele2[i], cpl_table_get_string(lineratio_calib_table,"ELEMENT2",i));
		calib[i][0] = cpl_table_get_double(lineratio_calib_table,"C0",i,NULL);
		calib[i][1] = cpl_table_get_double(lineratio_calib_table,"C1",i,NULL);
		calib[i][2] = cpl_table_get_double(lineratio_calib_table,"C2",i,NULL);
		calib[i][3] = cpl_table_get_double(lineratio_calib_table,"C3",i,NULL);
		nfit[i] =  cpl_table_get_int(lineratio_calib_table,"FIT_TYPE",i,NULL);
	}

/* Getting the EWs*/
	cpl_size na = cpl_table_get_nrow(line_ew_table);
	double lares[na],ewares[na];

	for (i=0;i<na;i++)	{
		lares[i] = cpl_table_get_double(line_ew_table,"WAVEL",i,NULL);
		ewares[i] = cpl_table_get_double(line_ew_table,"EW",i,NULL);
	}

/*Calculation of the ratios:*/
	double lratios[nl];
	esp_compile_ratios(lares, ewares, na, nl, line1, line2, lratios);
	double teff[nl],teff1[nl],teff2[nl],stddev1[nl],stddev2[nl];/*,lratios1[nl];*/
	char ele1_1[nl][6],ele1_2[nl][6],ele2_1[nl][6],ele2_2[nl][6],eleind1[nl][6],eleind2[nl][6];
	int n1=0;
	int n2=0;

	for(i=0;i<nl;i++) {
		if (nfit[i] >4)	{
			if (lratios[i] < 0) lratios[i]=0.001;
			teff[i]=calib[i][0]+calib[i][1]*log10(lratios[i])+calib[i][2]*log10(lratios[i])*log10(lratios[i])+calib[i][3]*log10(lratios[i])*log10(lratios[i])*log10(lratios[i]);
		} else {
      		if (nfit[i] >2) lratios[i]=1./lratios[i];
      		teff[i]=calib[i][0]+calib[i][1]*lratios[i]+calib[i][2]*lratios[i]*lratios[i]+calib[i][3]*lratios[i]*lratios[i]*lratios[i];
		}
		if (teff[i] >=4200. && teff[i] <=6800. && lratios[i] >=0.01 && lratios[i] <=100) {
			teff1[n1]=teff[i];
			stddev1[n1]=stddev[i];
/*			lratios1[n1]=lratios[i];*/
			strcpy (ele1_1[n1],ele1[i]);
			strcpy (ele2_1[n1],ele2[i]);
			n1++;
		}
	}

	double sigma1=0.,soma=0.,meanteff=0.;
	soma=0;
	for(i=0;i<n1;i++)	soma+=teff1[i];
	meanteff=soma/n1;
	soma=0;
	for(i=0;i<n1;i++)	soma+=(teff1[i]-meanteff)*(teff1[i]-meanteff);
	sigma1=sqrt (soma/(n1-1));

	for(i=0;i<n1;i++) {
		if (abs(teff1[i]-meanteff) <= 2.*sigma1 ) {
			teff2[n2]=teff1[i];
			stddev2[n2]=stddev1[i];
			strcpy (ele1_2[n2],ele1_1[i]);
			strcpy (ele2_2[n2],ele2_1[i]);
			n2++;
		}
	}

	double sigma2=0.,meanteff2=0.;
	soma=0;
	for(i=0;i<n2;i++)	soma+=teff2[i];
	meanteff2=soma/n2;
	soma=0;
	for(i=0;i<n2;i++)	soma+=(teff2[i]-meanteff2)*(teff2[i]-meanteff2);
	sigma2=sqrt (soma/(n2-1));

	soma=0;
	double soma2=0;
	for(i=0;i<n2;i++) {
		soma+=teff2[i]/(stddev2[i]*stddev2[i]);
		soma2+=1./(stddev2[i]*stddev2[i]);
	}
	double Teff_final,erTeff;
	int flag;
	int nrind=1,j;
	strcpy (eleind1[0],ele1_2[0]);
	strcpy (eleind2[0],ele2_2[0]);

	for(i=0;i<n2;i++) {
		flag=0;
		for (j=0;j<nrind;j++) {
			if ( (strcmp(eleind1[j],ele1_2[i]) == 0 && strcmp(eleind2[j],ele2_2[i]) == 0) || (strcmp(eleind1[j],ele2_2[i]) == 0 && strcmp(eleind2[j],ele1_2[i]) ==0 )) {
				flag=1;
			}
		}
		if (flag == 0) {
			strcpy (eleind1[nrind],ele1_2[i]);
			strcpy (eleind2[nrind],ele2_2[i]);
			nrind++;
		}
	}

	/*
	Correction for stars hotter than 6000 K
	(teffvecw(hteff)-res(0))/res(1)
	res0       3191.1655
	res1      0.47104218
	*/


	Teff_final=soma/soma2;
	erTeff=sigma2;
	double Teff_final_cor;
	if (Teff_final > 6000) Teff_final_cor=(Teff_final-3191.1655)/0.47104218; else Teff_final_cor=Teff_final;

	if (Teff_final > 6000) {	
		cpl_msg_info(cpl_func, "*** Correction for stars hooter than 6000 K considered***");
		cpl_msg_info(cpl_func, "*** Teff Corrected: %f +- %f ***",Teff_final_cor,erTeff/sqrt(nrind));
		
		if (Teff_final_cor < 6000) {
//			printf("Correction invalid for this star because Teff_cor < 6000 K.\n");
			cpl_msg_info(cpl_func, "*** Correction invalid for this star because Teff_cor < 6000 K. ***");
			Teff_final_cor=Teff_final;
		}
	}
	cpl_msg_info(cpl_func,"*** Final Teff: %6.1f +- %5.1f K. ***",Teff_final_cor,erTeff/sqrt(nrind));

	cpl_ensure_code(espda_create_teff_table(teff_table, 1)
                    == CPL_ERROR_NONE, cpl_error_get_code());

    cpl_table_set_double(*teff_table,"TEFF",0,Teff_final_cor);
    cpl_table_set_double(*teff_table,"STD_TEFF",0,erTeff);
    cpl_table_set_int(*teff_table,"N_CALIB",0,n2);
    cpl_table_set_int(*teff_table,"N_CALIB_IND",0,nrind);



  if (!cpl_errorstate_is_equal(prestate)) {
      return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
             "Unable to Get region of the spectrum");
  }


  return CPL_ERROR_NONE;

}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the value of the [Fe/H] ratio
  @param    linefeh_calib_table  cpl table with the calibrations for each ratio
  @param    line_ew_table	     cpl table with the EW measurements for the lines
  @param    teff_table      	 cpl table with the computed teff
  @param    feh_table      	     cpl table to store the computed [Fe/H]
  @return   cpl error code of execution

  This function computes the [Fe/H] for Fe I line. Computes the average value 
  removing some outliers and return the resulting [Fe/H], its standard deviation,
  and the number of lines used in the computation. (check Sousa et al. 2012A&A...544A.122S).

*/
/*------------------------------------------------------------------------------*/

cpl_error_code espda_calib_feh(cpl_table* linefeh_calib_table, 
							 cpl_table* line_ew_table,
							 cpl_table* teff_table, 
							 cpl_table** feh_table) {
							 
	cpl_errorstate prestate = cpl_errorstate_get();

	cpl_size nl = cpl_table_get_nrow(linefeh_calib_table);
	int i, j;
	/*int ni[nl],nstars[nl];*/
	double stddev[nl],line[nl],teffa[nl],teffb[nl],feha[nl],fehb[nl],ewa[nl],ewb[nl];
	char ele[nl][6];
	double calib[nl][6];


	for (i=0;i<nl;i++) {
    	stddev[i] = cpl_table_get_double(linefeh_calib_table,"STDDEV_FIN",i,NULL);
		line[i] = cpl_table_get_double(linefeh_calib_table,"WAVE",i,NULL);
		calib[i][0] = cpl_table_get_double(linefeh_calib_table,"C0",i,NULL);
		calib[i][1] = cpl_table_get_double(linefeh_calib_table,"C1",i,NULL);
		calib[i][2] = cpl_table_get_double(linefeh_calib_table,"C2",i,NULL);
		calib[i][3] = cpl_table_get_double(linefeh_calib_table,"C3",i,NULL);
		calib[i][4] = cpl_table_get_double(linefeh_calib_table,"C4",i,NULL);
		calib[i][5] = cpl_table_get_double(linefeh_calib_table,"C5",i,NULL);
    	teffa[i] = cpl_table_get_double(linefeh_calib_table,"TEFF_LOW",i,NULL);
    	teffb[i] = cpl_table_get_double(linefeh_calib_table,"TEFF_UP",i,NULL);
    	feha[i] = cpl_table_get_double(linefeh_calib_table,"FEH_LOW",i,NULL);
    	fehb[i] = cpl_table_get_double(linefeh_calib_table,"FEH_UP",i,NULL);
    	ewa[i] = cpl_table_get_double(linefeh_calib_table,"EW_LOW",i,NULL);
    	ewb[i] = cpl_table_get_double(linefeh_calib_table,"EW_UP",i,NULL);
	}

/* Getting the EWs*/
	cpl_size na = cpl_table_get_nrow(line_ew_table);
	double lares[na],ewares[na];

	for (i=0;i<na;i++)	{
		lares[i] = cpl_table_get_double(line_ew_table,"WAVEL",i,NULL);
		ewares[i] = cpl_table_get_double(line_ew_table,"EW",i,NULL);
	}

/* Getting the Teff*/
	double teff = cpl_table_get_double(teff_table,"TEFF",0,NULL);

/*Get EW for all lines that are present...*/
	double ewaresin[nl];
	for (i=0;i<nl;i++) {
		ewaresin[i]=-1;
		for (j=0;j<na;j++) {
			if (fabs(lares[j]-line[i]) < 0.01){
				ewaresin[i]=ewares[j];
			}
		}
	}

/*Compute FE/H:*/
	double feh[nl],feh1[nl],feh2[nl],stddev1[nl],stddev2[nl];
	int n1=0;
	int n2=0;

	for(i=0;i<nl;i++) {
		double va;
		double vb;
		double vc;
		va= calib[i][3];
		vb= (calib[i][1] + calib[i][5] * teff);
		vc= (calib[i][0] + calib[i][2]*teff + calib[i][4]*teff*teff - ewaresin[i]);
		if (vb*vb < 4*va*vc)
			feh[i]=-101.;
		else
			feh[i]= (-vb + sqrt(vb*vb - 4*va*vc) )/ (2*va) ;
		if ( (ewaresin[i] > ewa[i]-0.  && ewaresin[i] < ewb[i]+0.) && (teff >= teffa[i]-100.  && teff <= teffb[i]+100. ) &&  ( feh[i] >= feha[i]-0. && feh[i] <= fehb[i]+0.  )   )	{
			feh1[n1]=feh[i];
			stddev1[n1]=stddev[i];
			n1++;
		}
	}
	double sigma1=0.,soma=0.,meanfeh=0.;
	soma=0;
	for(i=0;i<n1;i++)	soma+=feh1[i];
	meanfeh=soma/n1;
	soma=0;
	for(i=0;i<n1;i++)	soma+=(feh1[i]-meanfeh)*(feh1[i]-meanfeh);
	sigma1=sqrt (soma/(n1-1));

	for(i=0;i<n1;i++) {
		if (abs(feh1[i]-meanfeh) <= 2.*sigma1) {
			feh2[n2]=feh1[i];
			stddev2[n2]=stddev1[i];
			n2++;
		}
	}
  
	double sigma2=0.,meanfeh2=0.;
	soma=0;
	for(i=0;i<n2;i++)	soma+=feh2[i];
	meanfeh2=soma/n2;
	soma=0;
	for(i=0;i<n2;i++)	soma+=(feh2[i]-meanfeh2)*(feh2[i]-meanfeh2);
	sigma2=sqrt (soma/(n2-1));

	soma=0;
	double soma2=0;
	for(i=0;i<n2;i++) {
		soma+=feh2[i]/(stddev2[i]*stddev2[i]);
		soma2+=1./(stddev2[i]*stddev2[i]);
	}

	double Feh_final,erFeh;
	Feh_final=soma/soma2;
	erFeh=sigma2;

	/*
	*fehout=Feh_final;
	*erfehout=erFeh;
	*nout=n2;
	*/	

	cpl_msg_info(cpl_func,"*** Final [Fe/H]: %6.3f +- %5.3f dex. ***",Feh_final,erFeh);

    cpl_ensure_code(espda_create_feh_table(feh_table, 1)
                    == CPL_ERROR_NONE, cpl_error_get_code());

    cpl_table_set_double(*feh_table,"FEH",0,Feh_final);
    cpl_table_set_double(*feh_table,"STD_FEH",0,erFeh);
    cpl_table_set_int(*feh_table,"N_CALIB",0,n2);



  if (!cpl_errorstate_is_equal(prestate)) {
      return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
             "Unable to Get region of the spectrum");
  }


  return CPL_ERROR_NONE;
}




/*----------------------------------------------------------------------------*/
/**
  @brief    Create a TEFF table with a given row size (1 default)
  @param    tab  New table
  @param    size Row size
  @return   CPL_ERROR_NONE ff OK

        
  This functions creates a CPL table with structure TEFF and a given row size.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_create_teff_table(cpl_table **tab,
                                       cpl_size    size) 

{
    cpl_errorstate prestate = cpl_errorstate_get();
    cpl_size       ceil;

/*
 * Table of given row size is created. Columns are those specified for the 
 * structure SPEC.
 */

    *tab = cpl_table_new(size);
    cpl_table_new_column(*tab, "TEFF",       CPL_TYPE_DOUBLE);
    cpl_table_new_column(*tab, "STD_TEFF",   CPL_TYPE_DOUBLE);
    cpl_table_new_column(*tab, "N_CALIB",    CPL_TYPE_INT);
    cpl_table_new_column(*tab, "N_CALIB_IND",CPL_TYPE_INT);

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to create table.");
    }

/*
 * Columns are initialized with default values.
 */

    if (size > 0) {
        ceil = size;
    }
    else {
        ceil = 0;
    }
    cpl_table_fill_column_window_double(*tab, "TEFF",        0, ceil, -9999.0);
    cpl_table_fill_column_window_double(*tab, "STD_TEFF",    0, ceil, -9999.0);
    cpl_table_fill_column_window_int(*tab, "N_CALIB",     0, ceil,    -9999);
    cpl_table_fill_column_window_int(*tab, "N_CALIB_IND", 0, ceil,    -9999);

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to initialize table.");
    }

/* 
 * The new table is passed along. 
 */

    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Create a FEH table with a given row size (1 default)
  @param    tab  New table
  @param    size Row size
  @return   CPL_ERROR_NONE ff OK

        
  This functions creates a CPL table with structure TEFF and a given row size.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_create_feh_table(cpl_table **tab,
                                      cpl_size    size) 

{
    cpl_errorstate prestate = cpl_errorstate_get();
    cpl_size       ceil;

/*
 * Table of given row size is created. Columns are those specified for the 
 * structure SPEC.
 */

    *tab = cpl_table_new(size);
    cpl_table_new_column(*tab, "FEH",       CPL_TYPE_DOUBLE);
    cpl_table_new_column(*tab, "STD_FEH",   CPL_TYPE_DOUBLE);
    cpl_table_new_column(*tab, "N_CALIB",   CPL_TYPE_INT);

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to create table.");
    }

/*
 * Columns are initialized with default values.
 */

    if (size > 0) {
        ceil = size;
    }
    else {
        ceil = 0;
    }
    cpl_table_fill_column_window_double(*tab, "FEH",        0, ceil, -9999.0);
    cpl_table_fill_column_window_double(*tab, "STD_FEH",    0, ceil, -9999.0);
    cpl_table_fill_column_window_int(*tab, "N_CALIB",     0, ceil,    -9999);

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to initialize table.");
    }

/* 
 * The new table is passed along. 
 */

    return CPL_ERROR_NONE;
}


/**@}*/
