/* 
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Id: espda_voigt.c 173006 2015-05-06 15:44:32Z gcupani $
 * $Author: gcupani $
 * $Date: 2015-05-06 17:44:32 +0200 (Wed, 06 May 2015) $
 * $Revision: 173006 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_voigt.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup  espda_voigt  Voigt parameters
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief   Create data structures with Voigt parameters.
  @param   line       A list of lines (LINE_VOIGT).
  @param   par_pos    The positions of varying and fixed parameters.
  @param   par_fix    The fixed parameters.
  @param   par_vary   The varying parameters.
  @param   par_lim    The limits on parameters.
  @param   vary_num   Number of varying parameters.
  @return  CPL error code.

  This function takes a list @a line containing @f$ L @f$ lines and copies their 
  Voigt parameters into data structures suitable for fitting. The Voigt 
  parameters are taken from columns @a LINEZ, @a COLDEN, @a THERB, and @a TURBB 
  of @a line; their range and constraints are taken for each parameter from 
  columns @a *MIN/@a *MAX and @a *CON. The following structures are produced:

    - @a par_pos: a @f$ 4L @f$-dimensional integer array to relate the positions 
      of the Voigt parameters in line to their positions in @a par_fix and 
      @a par_vary (see below). The rules are arcane; this parameter is not meant 
      to be handled by the user.
    - @a par_fix: a @f$ F @f$-dimensional double array with parameters to be 
      kept fixed. These are the @f$ F @f$ Voigt parameters with @a *CON equal to 
      "F".
    - @a par_vary: a @f$ (4L-F-C) @f$-dimensional double array with the 
      parameters allowed to vary. These are the Voigt parameters with @a *CON 
      not equal to "V". Their number @a vary_num is also issued. @f$ C @f$ is 
      the number of mutual constraints between parameters; if two parameters are 
      constrained to be equal in the fitting procedure, they are counted as one 
      parameter in @a par_vary.
    - @a par_lim: a @f$ 4L @f$-row table with minimum and maximum value for each 
      Voigt parameter. The values are saved in columns @a MIN, @a MAX and 
      duplicate in columns @a MIN_FIX, @a MAX_FIX (@a MIN_VARY, @a MAX_VARY) for 
      fixed (varying) parameters.
  
  Depending on the value of the global variable @a TRANSF_MODE, a transformation 
  may be applied to the parameters to enforce the prescribed ranges. For details 
  on the transformation, see @p espda_voigt_transf.
 
  If @a line is NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_voigt_create(cpl_table   *line,
                                  cpl_array  **par_pos,
                                  cpl_array  **par_fix,
                                  gsl_vector **par_vary,
                                  cpl_table  **par_lim,
                                  cpl_size    *vary_num)

{
    cpl_errorstate prestate = cpl_errorstate_get();

    cpl_size par_size  = 4 * cpl_table_get_nrow(line);
    *par_pos = cpl_array_new(par_size, CPL_TYPE_INT);
    *par_fix = cpl_array_new(par_size, CPL_TYPE_DOUBLE);
    *par_lim = cpl_table_new(par_size);
    cpl_table_new_column(*par_lim, "MIN", CPL_TYPE_DOUBLE);
    cpl_table_new_column(*par_lim, "MAX", CPL_TYPE_DOUBLE);
    cpl_table_new_column(*par_lim, "MIN_FIX", CPL_TYPE_DOUBLE);
    cpl_table_new_column(*par_lim, "MAX_FIX", CPL_TYPE_DOUBLE);
    cpl_table_new_column(*par_lim, "MIN_VARY", CPL_TYPE_DOUBLE);
    cpl_table_new_column(*par_lim, "MAX_VARY", CPL_TYPE_DOUBLE);

    cpl_array *par_temp      = cpl_array_new(par_size, CPL_TYPE_DOUBLE);
    cpl_array *par_fix_temp  = cpl_array_new(par_size, CPL_TYPE_DOUBLE);
    cpl_array *par_vary_temp = cpl_array_new(par_size, CPL_TYPE_DOUBLE);
    cpl_table *par_con  = cpl_table_new(0);
    cpl_table_new_column(par_con, "CON", CPL_TYPE_STRING);
    cpl_table_new_column(par_con, "POS", CPL_TYPE_INT);
    int j_fix = 0, j_vary = 0;

    // Loop over lines
    for (int i = 0, j = 0; i < (int)par_size / 4; i++) { 

        // Loop over parameters
        for (int k = 0; k < 4; k++, j++) {
            char   par_col[20], con_col[20], min_col[20], max_col[20];

            // Initialize the column names for the parameters and their errors
            if (j % 4 == 0) {
                strcpy(par_col, "LINEZ");
            }
            if (j % 4 == 1) {
                strcpy(par_col, "COLDEN");
            }
            if (j % 4 == 2) {
                strcpy(par_col, "THERB");
            }
            if (j % 4 == 3) {
                strcpy(par_col, "TURBB");
            }
            strcpy(con_col, par_col);
            strcpy(min_col, par_col);
            strcpy(max_col, par_col);
            strcat(con_col, "CON");
            strcat(min_col, "MIN");
            strcat(max_col, "MAX");

            // Copy parameters from the line list
            double  par = cpl_table_get_double(line, par_col, i, NULL);
            double  min, max;
            double  par_proc, err_proc;
            
            min = cpl_table_get_double(line, min_col, i, NULL);
            max = cpl_table_get_double(line, max_col, i, NULL);

            cpl_ensure_code(espda_voigt_transf(par, 0, min, max, "TRANSF", 
                &par_proc, &err_proc) == CPL_ERROR_NONE, cpl_error_get_code());
            
            cpl_table_set_double(*par_lim, "MIN", 4 * i + k, min);
            cpl_table_set_double(*par_lim, "MAX", 4 * i + k, max);

            // Depending on constraints, define parameters as fixed or varying
            // and construct the position array accordingly (varying [fixed]
            // parameters have positive [negative] position indexes starting 
            // from 0 [-1])
            const char *con = cpl_table_get_string(line, con_col, i);
            int         pos;
            if (!strcmp(con, "F")) {
                /*
                min = cpl_table_get_double(line, min_col, i, NULL);
                max = cpl_table_get_double(line, max_col, i, NULL);
                cpl_ensure_code(espda_voigt_transf(par, 0, min, max, "TRANSF", 
                    &par_proc, &err_proc) == CPL_ERROR_NONE, cpl_error_get_code());
                */
                cpl_array_set_double(*par_fix, j_fix, par_proc);

                cpl_array_set_double(par_fix_temp, j_fix, par);
                cpl_table_set_double(*par_lim, "MIN_FIX", j_fix, min);
                cpl_table_set_double(*par_lim, "MAX_FIX", j_fix, max);
                
                j_fix++;
                pos = -j_fix;
            }
            else if (!strcmp(con, "V")) {
                /*
                min = cpl_table_get_double(line, min_col, i, NULL);
                max = cpl_table_get_double(line, max_col, i, NULL);
                cpl_ensure_code(espda_voigt_transf(par, 0, min, max, "TRANSF", 
                    &par_proc, &err_proc) == CPL_ERROR_NONE, cpl_error_get_code());
                */
                cpl_array_set_double(par_temp, j_vary, par_proc);

                cpl_array_set_double(par_vary_temp, j_vary, par);
                cpl_table_set_double(*par_lim, "MIN_VARY", j_vary, min);
                cpl_table_set_double(*par_lim, "MAX_VARY", j_vary, max);

                pos =  j_vary;
                j_vary++;
            }
            else {
                cpl_table_unselect_all(par_con);
                cpl_table_or_selected_string(par_con, "CON", CPL_EQUAL_TO, con);
                cpl_table *con_ref = cpl_table_extract_selected(par_con);
                if (cpl_table_get_nrow(con_ref) == 0) {
                    cpl_table_set_size  (par_con, 
                        cpl_table_get_nrow(par_con) + 1);
                    cpl_table_set_string(par_con, "CON", 
                        cpl_table_get_nrow(par_con) - 1, con);
                    cpl_table_set_int   (par_con, "POS", 
                        cpl_table_get_nrow(par_con) - 1, j_vary);
                    /*                    
                    min = cpl_table_get_double(line, min_col, j_vary, NULL);
                    max = cpl_table_get_double(line, max_col, j_vary, NULL);
                    cpl_ensure_code(espda_voigt_transf(par, 0, min, max, "TRANSF", 
                        &par_proc, &err_proc) == CPL_ERROR_NONE, cpl_error_get_code());
                    */
                    cpl_array_set_double(par_temp, j_vary, par_proc);

                    cpl_array_set_double(par_vary_temp, j_vary, par);
                    cpl_table_set_double(*par_lim, "MIN_VARY", j_vary, min);
                    cpl_table_set_double(*par_lim, "MAX_VARY", j_vary, max);
                    
                    pos = j_vary;
                    j_vary++;
                }
                else {
                    pos = cpl_table_get_int(con_ref, "POS", 0, NULL);
                    cpl_table_set_double(line, min_col, i, 
                        cpl_table_get_double(*par_lim, "MIN_VARY", pos, NULL));
                    cpl_table_set_double(line, max_col, i, 
                        cpl_table_get_double(*par_lim, "MAX_VARY", pos, NULL));
                }
                cpl_table_delete(con_ref);
            }
            cpl_array_set_int(*par_pos, j, pos);
        }
    }
    cpl_table_delete(par_con);
    
    if (j_fix > 0) {
        cpl_array_erase_window(*par_fix, j_fix,  par_size - j_fix);
        cpl_array_erase_window(par_temp, j_vary, par_size - j_vary);
    }
    *vary_num = j_vary;
    *par_vary = gsl_vector_alloc(*vary_num);
    for (int j = 0; j < *vary_num; j++) {
        gsl_vector_set(*par_vary, j, cpl_array_get_double(par_temp, j, NULL));
    }
    //cpl_array_dump(*par_pos, 0, 100, NULL);
    //cpl_array_dump(par_temp, 0, 100, NULL);
    cpl_array_delete(par_temp);

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to import Voigt parameters.");
    }
    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the integrand of the Voigt function.
  @param    x  The argument of the Voigt function.
  @param    p  The parameters of the Voigt function.
  @return   The integrand.

  This function computes the integrand which appears in the representation of 
  the Voigt function described by Zaghloul 2007, as a function of the argument 
  @a x and the parameters @a p.

  The integrand is @f[
    I(x;a,u)=\mathrm{exp}\left[- \left(u^2-x^2\right)\right]
             \mathrm{sin}\left[2a\left(u  -x  \right)\right],
  @f]
  where @f$ a,u @f$ are the two parameters in @a p.
  
  N.B. This function is meant to be called only by the @p gsl_integration 
       package.
 */
/*----------------------------------------------------------------------------*/

double espda_voigt_integ(double  x,
                         void   *p)

{
    cpl_errorstate prestate = cpl_errorstate_get();

    struct espda_voigt_integ_params *params = 
        (struct espda_voigt_integ_params *)p;
    double a = (params->a);
    double u = (params->u);

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to compute the Voigt function integrand.");
    }
    return exp(-(u * u - x * x)) * sin(2 * a * (u - x));
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Merge data structures with Voigt parameters
  @param    lfit       Line list (LFIT)
  @param    ion_wavel  Wavelengths of the ionic transitions
  @param    par_pos    Positions of varying and fixed parameters
  @param    par_fix    Fixed parameters
  @param    par_vary   Varying parameters
  @param    err_vary   Errors on varying parameters
  @param    par_lim    Limits on parameters
  @param    par_all    All parameters
  @param    err_all    All errors on parameters
  @return   CPL_ERROR_NONE iff OK  

  This function merges the data structures with the varying and fixed line
  parameters used for Voigt profile fitting. A single array with all line 
  parameters is needed when computing the Voigt function. Parameters are copied
  from the data structures containing them and optionally transformed back to
  their physical values. The function handles also the errors on varying 
  parameters (if they are provided). If a line list is given in input, it is
  updated with the current values of parameters and (optionally) their errors.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_voigt_merge(cpl_table   	  *lfit,
                                 cpl_array   	  *ion_wavel,
                                 cpl_array   	  *par_pos,
                                 cpl_array   	  *par_fix,
                                 const gsl_vector *par_vary,
                                 const gsl_matrix *err_vary,
                                 cpl_table  	  *par_lim,
                                 cpl_array  	 **par_all,
                                 cpl_array  	 **err_all)

{
    cpl_errorstate prestate = cpl_errorstate_get();
    cpl_size par_size  = cpl_array_get_size(par_pos);
    espda_struct_check(par_all, *par_all = cpl_array_new(par_size, 
        CPL_TYPE_DOUBLE));
    espda_struct_check(err_all, *err_all = cpl_array_new(par_size, 
        CPL_TYPE_DOUBLE));
    
    // Loop over lines
    for (int i = 0, j = 0; i < (int)par_size / 4; i++) { 

        // Loop over parameters
        for (int k = 0; k < 4; k++, j++) {
            int    pos = cpl_array_get_int(par_pos, j, NULL); 
            double err = 0.0, par;
            double min, max;
            double par_proc, err_proc;
            if (pos < 0) {
                par = cpl_array_get_double(par_fix, -pos - 1, NULL);
                
                min = cpl_table_get_double(par_lim, "MIN_FIX", -pos - 1, NULL);
                max = cpl_table_get_double(par_lim, "MAX_FIX", -pos - 1, NULL);
                
            } 
            else {
                par = gsl_vector_get(par_vary, pos);

                min = cpl_table_get_double(par_lim, "MIN_VARY", pos, NULL);
                max = cpl_table_get_double(par_lim, "MAX_VARY", pos, NULL);

                espda_struct_check(err_vary, err = sqrt(gsl_matrix_get(err_vary, 
                    pos, pos)));
            }
            cpl_ensure_code(espda_voigt_transf(par, err, min, max, "ATRANSF", 
                &par_proc, &err_proc) == CPL_ERROR_NONE, cpl_error_get_code());
            
            // Copy parameters and errors to new arrays and to line list
            espda_struct_check(par_all, cpl_array_set_double(*par_all, j, 
                par_proc));
            espda_struct_check(err_all, cpl_array_set_double(*err_all, j, 
                err_proc));

            // Initialize the column names for the parameters and their errors
            if (lfit != NULL) {
	        char   par_col[20], err_col[20];
                if (j % 4 == 0) {
                    strcpy(par_col, "LINEZ");
                }
                if (j % 4 == 1) {
                    strcpy(par_col, "COLDEN");
                }
                if (j % 4 == 2) {
                    strcpy(par_col, "THERB");
                }
                if (j % 4 == 3) {
                    strcpy(par_col, "TURBB");
                }
                strcpy(err_col, par_col);
                strcat(err_col, "ERR");
                cpl_table_set_double(lfit, par_col, i, par_proc);
                cpl_table_set_double(lfit, err_col, i, err_proc);
                if (j % 4 == 0) {
                    cpl_table_set_double(lfit, "WAVEL", i, (par_proc + 1.0)  *
                        cpl_array_get_double(ion_wavel, i, NULL));
                    cpl_table_set_double(lfit, "WAVELERR", i, err_proc *
                        cpl_array_get_double(ion_wavel, i, NULL));
                }
                if (j % 4 == 2) {
                    double wavel = cpl_table_get_double(lfit, "WAVEL", i, NULL);
		    //FWHM
                    double hwidth = 2.35482 * par_proc / LIGHTSPEED * wavel;
                    cpl_table_set_double(lfit, "WAVELMIN", i, wavel - hwidth);
                    cpl_table_set_double(lfit, "WAVELMAX", i, wavel + hwidth);
                }
            }
        }
    }
    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to export Voigt parameters.");
    }
    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Prepare a line list for Voigt profile fitting
  @param    line          Starting line list (LINE)
  @param    line_add      Optional added line list (LINE)
  @param    ion           Reference list of ionic transitions (ION)
  @param    hwidth        Halfwidth for line grouping [nm] 
  @param    zem           Emission redshift
  @param    verbose       Verbosity flag
  @param    par_range     Range of line parameters, coded as a string
  @param    par_edit      Changes to line parameters w.r.t. default, coded
                          as a string
  @param    line_pre          Output line list (LFIT)
  @return   CPL_ERROR_NONE iff OK  

  This function takes a list of absorption lines and fills it with guess 
  parameters for Voigt profile fitting. An optional list of manually added lines
  is optionally merged to the starting list. Parameter ranges and values can be
  adjusted using par_range and par_edit:
 
  - par_range must be formatted as a string of 7 comma-separated values, which
    denote, in order: [0] half width of the region where the line position may
    vary (in units of redshift); [1]-[2] minimum and maximum value column 
    density (in logaritmic units); [3]-[4] minimum and maximum thermal
    broadening; [5]-[6] minimum and maximum turbulence broadening. This value
    are used to fill the columns [NAME]MIN and [NAME]MAX of the line list, where
    [NAME] stands LINEZ, COLDEN, THERB, or TURBB.
  - par_edit must be formatted as a string of semicolon-separated instructions,
    each one given as follows: [line number],[parameter name],[parameter value]
    (e.g. "0,COLDEN,15;1,THERB,30;"). Parameters that can be edited are [NAME], 
    [NAME]MIN, [NAME]MAX, and [NAME]CON, where, as before, [NAME] stands for  
    LINEZ, COLDEN, THERB, or TURBB. [NAME]CON must be "V" for varying 
    parameters, "F" for fixed parameters, and any other string for common  
    parameters (all common parameters must have the same string).
  .      

  Lines are assigned to groups according to their mutual separation and to the 
  mutual constraints on line parameters. The group ID are listed in column GRPID
  of the output list. Lines are grouped together if their WAVELMIN to WAVELMAX
  ranges overlap, or if they are joined by at list one common parameter.
  
  If the emission redshift is different from zero, the line list is interpreted 
  as a quasar one and the line redshifts are computed using a reference list of 
  ionic transitions. A warning is issued of a line redshift is higher than the 
  quasar emission redshift.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_voigt_prep(cpl_table *line, cpl_table *ion, double hwidth, 
    double zem, const char *par_range, cpl_table  **line_pre) {
    extern int v_level;    
    cpl_errorstate prestate = cpl_errorstate_get();

    if (line != NULL) {

        // Set line parameter ranges from command line parameter
        double par_range_value[7];
        if (strcmp(par_range, INV_STRING) != 0) {
            char   range[LEN_STRING], *null = NULL;
            strcpy(range, par_range);
            char  *range_token = strtok(range, ",");
            int i = 0; 
            while (range_token != NULL) {
                par_range_value[i] = strtod(range_token, &null);
                range_token = strtok(NULL, ",");
                i++;
            }
            if (i != 7) {
                cpl_msg_warning(cpl_func, "Parameter "
                    "ranges not properly formatted! Using default value "
                    "--par-range='1.0e-4,10.0,22.0,2.0,100.0,0.0,0.0'");
                par_range_value[0] = 1.0e-4;
                par_range_value[1] = 10.0;
                par_range_value[2] = 22.0;
                par_range_value[3] = 2.0;
                par_range_value[4] = 100.0;
                par_range_value[5] = 0.0;
                par_range_value[6] = 0.0;            
            }
        }

        // The whole table is prepared, if it is empty
        if (cpl_table_has_column(line, "COLDEN")) { 
            if (cpl_table_get_double(line, "COLDEN", 0, NULL) != INV_DOUBLE) {
                *line_pre = cpl_table_duplicate(line);
            }
        }
        if (*line_pre == NULL || 
            cpl_table_get_double(line, "COLDENERR", 0, NULL) == 0.0) { 
            espda_util_func(espda_line_fill(line, hwidth, zem, 
                par_range_value, line_pre), LINE_FILL, 0);
        }

        // The portion of the table with invalid values is prepared
        cpl_table_unselect_all(*line_pre);
        cpl_table_or_selected_invalid(*line_pre, "LINEZ");
        cpl_table *line_inv = cpl_table_extract_selected(*line_pre);
        cpl_table_select_all(*line_pre);
        cpl_table_erase_invalid(*line_pre);
        cpl_table *line_inv_pre = NULL;
        espda_util_safe(line_inv, espda_util_func(espda_line_fill(line_inv,
            hwidth, zem, par_range_value, &line_inv_pre), LINE_FILL, 0));
        espda_util_safe(line_inv_pre, cpl_table_insert(*line_pre, line_inv_pre, 
            cpl_table_get_nrow(*line_pre)));

        espda_struct_check(*line_pre, espda_struct_sort_old(*line_pre, "WAVEL", 
            FALSE));

        // Group lines
        espda_util_func(espda_line_group(*line_pre, ion, //zem, 
            par_range_value), LINE_GROUP, 0);
    }

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to prepare line list for Voigt profile fitting.");
    }
    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Trasform Voigt parameters
  @param    par        Parameter
  @param    err        Error
  @param    min        Parameter range minimum
  @param    max        Parameter range maximum
  @param    dir        "Direction" of transform
  @param    par_proc   Processed parameter
  @param    err_proc   Processed error
  @return   CPL_ERROR_NONE iff OK  

  This function is used to transform Voigt parameter before and after the 
  fitting procedure. Transform is aimed to make the parameter vary within a 
  limited range and to make variation of different parameters comparable. Two
  transform modes are available (defined in espda_dfs.h), "TAN" and "ASIN", 
  using respectively a tangent and and arcsine function; "NONE" denotes no
  transform. A parameter is transformed before the fitting procedure by setting
  dir to "TRANSF" and anti-transformed after the fitting procedure by setting
  dir to "ATRANSF". In this second case, the error on the parameter is also
  anti-transformed (obviously, this can be done only after the fitting 
  procedure).
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_voigt_transf(double        par,
                                  double        err,
                                  const double  min,
                                  const double  max,
                                  const char   *dir,
                                  double       *par_proc,
                                  double       *err_proc) 

{ 
    cpl_errorstate prestate = cpl_errorstate_get();

    *par_proc = par;
    *err_proc = err;        
    if (min * max != 0 && max != min) {
        if (!strcmp(TRANSF_MODE, "T"))  {  // Tangent
            if (!strcmp(dir, "TRANSF")) {
                *par_proc = tan((par - (max + min) * 0.5) * CPL_MATH_PI / 
                                 (max - min));
            }
            if (!strcmp(dir, "ATRANSF")) {
                if (err != 0.0) {
                    *err_proc = (1 / (1 + pow(par, 2.0))) * (max - min) *  
                                CPL_MATH_1_PI * err;
                }
                *par_proc = atan(par) * (max - min) * CPL_MATH_1_PI + 
                                  (max + min) * 0.5;
            }
        }
        else if (!strcmp(TRANSF_MODE, "A")) {  // Arc sine
            if (!strcmp(dir, "TRANSF")) {
                *par_proc = asin(2.0 * (par - min) / (max - min) - 1.0);
            }
            if (!strcmp(dir, "ATRANSF")) {
                if (err != 0.0) {
                    *err_proc = 0.5 *  cos(par) * (max - min) * err; 
                }
                *par_proc = 0.5 * (sin(par) + 1.0) * (max - min) + min;
            }
        }
    }
    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to tune Voigt parameters.");
    }
    return CPL_ERROR_NONE;
}

/**@}*/
