
/* $Id: espda_compu_eqwidth.c 270652 2022-02-25 08:58:39Z amodigli $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: amodigli $
 * $Date: 2022-02-25 08:58:39 +0000 (Fri, 25 Feb 2022) $
 * $Revision: 270652 $
 * $Name: espda_compu_eqwidth $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#include <string.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include "cpl_plot.h"

#include "espda_dfs.h"
#include "espda_eqwidth_lib.h"



#define LOC_SPEC_NPIX 1000


/*-----------------------------------------------------------------------------
                             Plugin registration
 -----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist * list);

/*-----------------------------------------------------------------------------
                            Private function prototypes
 -----------------------------------------------------------------------------*/

static int espda_compu_eqwidth_create(cpl_plugin *);
static int espda_compu_eqwidth_exec(cpl_plugin *);
static int espda_compu_eqwidth_destroy(cpl_plugin *);
static int espda_compu_eqwidth(cpl_frameset *, const cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char espda_compu_eqwidth_description[] =
"This recipe computes the equivalenth width for a given set of aborption lines"
"in a S1D spectra that is in a fits table."
"\n"
"INPUT FRAMES                        FORMAT  # EXPECTED  TAGS\n"
"1D stellar spectrum                 SPEC    1           RSPEC_PRE\n"
"list of absorption lines to measure LINE    1           LINE"
"\n"
"OUTPUT FRAMES                       FORMAT  # PRODUCED  TAGS                 "
"       NOTES\n"
"list of lines with EW measurements  LINE    1           LINE                 "
"      The input list of lines to be measured is updated and saved as new file"
;

/*-----------------------------------------------------------------------------
                                Function code
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok, 1 otherwise
  @note     Only this function is exported

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist *list)
{
    cpl_recipe *recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin *plugin = &recipe->interface;

    if (cpl_plugin_init(plugin, CPL_PLUGIN_API, ESPDA_BINARY_VERSION,
                        CPL_PLUGIN_TYPE_RECIPE, "espda_compu_eqwidth", "Compute"
                        " the equivalent widhts for a set of absorption lines",
                        espda_compu_eqwidth_description, "Sergio Sousa",
                        PACKAGE_BUGREPORT, espda_dfs_license(),
                        espda_compu_eqwidth_create, espda_compu_eqwidth_exec,
                        espda_compu_eqwidth_destroy)) {
        cpl_msg_error(cpl_func, "Plugin initialization failed");
        (void)cpl_error_set_where(cpl_func);
        return 1;
    }

    if (cpl_pluginlist_append(list, plugin)) {
        cpl_msg_error(cpl_func, "Error adding plugin to list");
        (void)cpl_error_set_where(cpl_func);
        return 1;
    }

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_eqwidth_create(cpl_plugin *plugin)
{
    cpl_recipe    *recipe;
    cpl_parameter *param;

    /* Do not create the recipe if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Parameter list allocation failed");
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
    }

    /* Fill the parameters list */
    /* --apply-mask-add */
/*
    param = cpl_parameter_new_value("espda.espda_mask_spec.mask_apply_add",
                                    CPL_TYPE_BOOL, "Flag to reset the mask "
                                    "column before applying the mask",
                                    "espda.espda_mask_spec", TRUE);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "apply-mask-add");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);
*/


    /* Fill the parameters list */

    /* -- wavel_min replaced ew_ll_begin*/
/*    param = cpl_parameter_new_value("espda.espda_compu_eqwidth.wavel_min",
                                    CPL_TYPE_DOUBLE, "Starting wavelength "
                                    "of the spectrum [nm]",
                                    "espda.espda_compu_eqwidth", 380.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "wavel_min");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* -- wavel_max replaced ew_ll_end*/
/*    param = cpl_parameter_new_value("espda.espda_compu_eqwidth.wavel_max",
                                    CPL_TYPE_DOUBLE, "Ending wavelength "
                                    "of the spectrum [nm]",
                                    "espda.espda_compu_eqwidth", 990.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "wavel_max");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* -- ew--space*/
    param = cpl_parameter_new_value("espda.espda_compu_eqwidth.space",
                                    CPL_TYPE_DOUBLE, "Width of the local"
                                    " spectral region for each side of \n"
                                    " the line [nm]; range: [0.1,5]. ",
                                    "espda.espda_compu_eqwidth", 0.3);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "space");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* -- ew_min replaced miniline */
    param = cpl_parameter_new_value("espda.espda_compu_eqwidth.ew_min",
                                    CPL_TYPE_DOUBLE, "Minimum accepted "
                                    "value for an EW to be saved in the product [mA];\n"
                                    " range [0,400]",
                                    "espda.espda_compu_eqwidth", 2.);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "ew_min");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* -- cont--rejt */
    param = cpl_parameter_new_value("espda.espda_compu_eqwidth.cont_rejt",
                                    CPL_TYPE_DOUBLE, "Rejection threshold"
                                    " for the continuum determination.\n"
                                    " The default value depends on the "
                                    "signal-to-noise (-1);\n"
                                    " range: [0.5,1[. ",
                                    "espda.espda_compu_eqwidth", -1.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "cont_rejt");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* -- cont--iter */
    param = cpl_parameter_new_value("espda.espda_compu_eqwidth.cont_iter",
                                    CPL_TYPE_INT, "Number of iterations"
                                    " to fit the continuum;\n"
                                    " range: [1,20]. ",
                                    "espda.espda_compu_eqwidth", 5);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "cont_iter");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* -- det--line--thres */
    param = cpl_parameter_new_value("espda.espda_compu_eqwidth.det_line_thres",
                                    CPL_TYPE_DOUBLE, "Normalized threshold for "
                                    "line acceptance in line identification;\n"
                                    " range: [0,0.5]. ",
                                    "espda.espda_compu_eqwidth", 0.02);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "det_line_thres");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);


    /* -- det--line---resol */
    param = cpl_parameter_new_value("espda.espda_compu_eqwidth.det_line_resol",
                                    CPL_TYPE_DOUBLE, "Minimum accepted separation"
                                    " between two lines \n used for the multi "
                                    " gaussian fits. It is also used to identify \n"
                                    " the lines to be measured [nm];"
                                    " range: [0.001,0.1]. ",
                                    "espda.espda_compu_eqwidth", 0.01);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "det_line_resol");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* -- det--line--smwidth */
    param = cpl_parameter_new_value("espda.espda_compu_eqwidth.det_line_smwidth",
                                    CPL_TYPE_INT, "Number of pixels in the boxcar"
                                    " to be averaged to fit the continuum;\n"
                                    " useful to deal with noisy spectrum; "
                                    " range: [1,50]. ",
                                    "espda.espda_compu_eqwidth", 4);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "det_line_smwidth");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* -- fit--ngauss--width */
    param = cpl_parameter_new_value("espda.espda_compu_eqwidth.fit_ngauss_width",
                                    CPL_TYPE_DOUBLE, "Initial guess for the "
                                    "parameter related with the"
                                    " width of the gaussians [mA];\n"
                                    " range: [0.1,2.5]. ",
                                    "espda.espda_compu_eqwidth", 0.5);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "fit_ngauss_width");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);



    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_eqwidth_exec(cpl_plugin *plugin)
{
    int             recipe_status;

    cpl_recipe     *recipe = NULL;
    cpl_errorstate  initial_errorstate = cpl_errorstate_get();

    /* Return immediately if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Verify parameter and frame lists */
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }
    if (recipe->frames == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Invoke the recipe */
    recipe_status = espda_compu_eqwidth(recipe->frames, recipe->parameters);

    /* Ensure DFS-compliance of the products */
    if (cpl_dfs_update_product_header(recipe->frames)) {
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();
    }

    if (!cpl_errorstate_is_equal(initial_errorstate)) {
        /* Dump the error history since recipe execution start.
           At this point the recipe cannot recover from the error */
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);
    }

    return recipe_status;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_eqwidth_destroy(cpl_plugin *plugin)
{
    cpl_recipe * recipe;

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    cpl_parameterlist_delete(recipe->parameters);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist    the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_eqwidth(cpl_frameset   *frameset,
			                   const cpl_parameterlist *paramlist)
{
//    cpl_errorstate    prestate = cpl_errorstate_get();
    cpl_table        *spec_table = NULL;
    cpl_table        *linelist_table = NULL;
    cpl_frame        *spec_frame = NULL;
    cpl_frame        *lines_frame = NULL;

// iteration cariables
    cpl_table* spec_region = NULL;
    cpl_table* line_table = NULL;
    cpl_table* spec_cont_region = NULL;
    cpl_table* spec_cont_region_der1 = NULL;
    cpl_table* spec_cont_region_der2 = NULL;
    cpl_table* spec_cont_region_der3 = NULL;
    double line;

    cpl_msg_info(cpl_func, "*** Entering Compu EQwidth ***");

    cpl_msg_info(cpl_func, "*** Checking 1D SPEC frame ***");


/*
 * DFS groups are set for all input frames and the data are loaded into a
 * merged table (for 1D spectra) and/or an order table (for 2D spectra).
 * Functions espda_dfs_groups() and espda_open_frames() are called.
 * Property list for QC parameters is created.
 */

    cpl_msg_info(cpl_func, "*** Set groups of FITS frames ***");
    cpl_ensure_code(espda_dfs_groups(frameset) == CPL_ERROR_NONE,
                    cpl_error_get_code());



// Getting the 1D spec fits table frame
    spec_frame = cpl_frameset_find(frameset, RSPEC_PRE);
// Checking if the MERGESPEC_TAB fits frame is defined
    if (spec_frame == NULL) {
        return (int)cpl_error_set_message(cpl_func,
               CPL_ERROR_DATA_NOT_FOUND, "The S1D RSPEC_PRE FITS"
               "TABLE is not defined: Need a frame with RSPEC_PRE");
    }

// Getting the keyword in the frame to check if is really a MERGESPEC
    cpl_propertylist* proplist = cpl_propertylist_load_regexp(
	                     cpl_frame_get_filename(spec_frame), 0, CPL_DFS_PRO_CATG, 0);
// Check if Keyword is present here:
    if (proplist == NULL) {
        return (int)cpl_error_set_message(cpl_func,
               CPL_ERROR_BAD_FILE_FORMAT, "The S1D RSPEC_PRE FITS"
               "File is not a RSPEC_PRE Frame - Keyword missing");
    }

// Check if Keyword is compatible
    if (strcmp(espda_dfs_catg(proplist),RSPEC_PRE)!=0){
        return (int)cpl_error_set_message(cpl_func,
               CPL_ERROR_BAD_FILE_FORMAT, "The S1D RSPEC_PRE FITS"
               "File is not a RSPEC_PRE Frame - Keyword not compatible");
    }
    cpl_propertylist_delete(proplist);
    cpl_msg_info(cpl_func, "*** 1D SPEC frame OK***");

// -------------------------------------------------------------------

    cpl_msg_info(cpl_func, "*** Checking LINE LIST TABLE frame ***");

// Getting the 1D spec fits table frame
    lines_frame = cpl_frameset_find(frameset, LINE);
// Checking if the LINE LIST fits frame is defined
    if (lines_frame == NULL) {
        return (int)cpl_error_set_message(cpl_func,
               CPL_ERROR_DATA_NOT_FOUND, "The LINE LIST FITS"
               "TABLE is not defined: Need a frame with LINE");
    }

// Getting the keyword in the frame to check if is really a LINE
    proplist = cpl_propertylist_load_regexp(
	                     cpl_frame_get_filename(lines_frame), 0, CPL_DFS_PRO_CATG, 0);

// Check if Keyword is present:
    if (proplist == NULL) {
        return (int)cpl_error_set_message(cpl_func,
               CPL_ERROR_BAD_FILE_FORMAT, "The LINE LIST FITS"
               "File is not a LINE Frame - Keyword missing");
    }

// Check if Keyword is compatible
    if (strcmp(espda_dfs_catg(proplist),LINE)!=0){
        return (int)cpl_error_set_message(cpl_func,
               CPL_ERROR_BAD_FILE_FORMAT, "The LINE LIST FITS"
               "File is not a LINE Frame - Keyword not compatible");
    }
    cpl_propertylist_delete(proplist);
    cpl_msg_info(cpl_func, "*** LINE LIST frame OK***");


    // Loading into a cpl table
    spec_table = cpl_table_load( cpl_frame_get_filename(spec_frame), 1, 0);


    // changing from Angstroms to Nanometers if needed
    int* error=NULL;
    double ll_init = cpl_table_get(spec_table, "WAVEL",0, error);
    if (ll_init/1000 > 1) {
        printf("Changing units from Angstroms to Nanometers in wavelength!\n");
        cpl_ensure_code(espda_change_wave_units(spec_table)
                    == CPL_ERROR_NONE, cpl_error_get_code());
    }


    printf("Filename Spec Fits Table: %s\n", cpl_frame_get_filename(spec_frame));
//    cpl_plot_column	("","","",spec_table,"WAVEL","FLUX");


    // printing parameters from the list
/*    printf("TEST1\n");
    espda_print_param(paramlist, "espda.espda_compu_eqwidth.ll_begin");
    printf("%s: %d\n", name, cpl_parameter_get_default_int(par));
    printf("TEST2\n");
//    print_param_double(paramlist, "espda.espda_compu_eqwidth.ll_end");
    print_param_double(paramlist, "espda.espda_compu_eqwidth.space");
    print_param_double(paramlist, "espda.espda_compu_eqwidth.miniline");
    print_param_double(paramlist, "espda.espda_compu_eqwidth.cont_rejt");
    print_param_int(paramlist, "espda.espda_compu_eqwidth.cont_iter");
    print_param_double(paramlist, "espda.espda_compu_eqwidth.det_line_thres");
    print_param_double(paramlist, "espda.espda_compu_eqwidth.det_line_resol");
    print_param_int(paramlist, "espda.espda_compu_eqwidth.det_line_smwidth");
    print_param_double(paramlist, "espda.espda_compu_eqwidth.fit_ngauss_width");
*/

/*    double wavel_min = cpl_parameter_get_double(
                      cpl_parameterlist_find_const(
                         paramlist,"espda.espda_compu_eqwidth.wavel_min"));
    printf("wavel_min: %lf\n", wavel_min);

    double wavel_max = cpl_parameter_get_double(
                      cpl_parameterlist_find_const(
                         paramlist,"espda.espda_compu_eqwidth.wavel_max"));
    printf("wavel_max: %lf\n", wavel_max);
*/    
    double wavel_min = 0;
    double wavel_max = 10000;

    double space = cpl_parameter_get_double(
                      cpl_parameterlist_find_const(
                         paramlist,"espda.espda_compu_eqwidth.space"));
    printf("space: %lf\n", space);

    double ew_min = cpl_parameter_get_double(
                      cpl_parameterlist_find_const(
                         paramlist,"espda.espda_compu_eqwidth.ew_min"));
    printf("ew_min: %lf\n", ew_min);
    // passage to nm
    ew_min /= 10.;


    if (space < 0.1 || space > 5) {
        return (int)cpl_error_set_message(cpl_func,
               CPL_ERROR_INCOMPATIBLE_INPUT, "-space- parameter out of bounds: [0.1,5]");
    }

    double cont_rejt = cpl_parameter_get_double(
                          cpl_parameterlist_find_const(
                             paramlist,"espda.espda_compu_eqwidth.cont_rejt"));
    printf("rejt: %lf\n", cont_rejt);

    if ((cont_rejt < 0.5 && cont_rejt != -1.0) || cont_rejt >=1.0 ) {
        	return (int)cpl_error_set_message(cpl_func,
               	CPL_ERROR_INCOMPATIBLE_INPUT, "-cont_rejt- parameter out of bounds: [0.5,1[");
    }

    int cont_iter = cpl_parameter_get_int(
                          cpl_parameterlist_find_const(
                             paramlist,"espda.espda_compu_eqwidth.cont_iter"));

    if (cont_iter < 1 || cont_iter >20 ) {
        return (int)cpl_error_set_message(cpl_func,
               CPL_ERROR_INCOMPATIBLE_INPUT, "-cont_iter- parameter out of bounds: [1,20]");
    }

    int det_line_smwidth = cpl_parameter_get_int(
                            cpl_parameterlist_find_const(
                             paramlist,"espda.espda_compu_eqwidth.det_line_smwidth"));

    if (det_line_smwidth < 1 || det_line_smwidth > 50 ) {
        return (int)cpl_error_set_message(cpl_func,
               CPL_ERROR_INCOMPATIBLE_INPUT, "-det_line_smwidth- parameter out of bounds: [1,50]");
    }

    double det_line_thres = cpl_parameter_get_double(
                             cpl_parameterlist_find_const(
                              paramlist,"espda.espda_compu_eqwidth.det_line_thres"));

    if (det_line_thres <= 0.0 || det_line_thres > 0.5 ) {
        return (int)cpl_error_set_message(cpl_func,
               CPL_ERROR_INCOMPATIBLE_INPUT, "-det_line_thres- parameter out of bounds: [0,0.5]");
    }

    double det_line_resol = cpl_parameter_get_double(
                             cpl_parameterlist_find_const(
                              paramlist,"espda.espda_compu_eqwidth.det_line_resol"));

    if (det_line_resol < 0.001 || det_line_resol > 0.1 ) {
        return (int)cpl_error_set_message(cpl_func,
               CPL_ERROR_INCOMPATIBLE_INPUT, "-det_line_resol- parameter out of bounds: [0.001,0.1]");
    }

// CAREFUL with this value for the correct gaussian fitting convergency
// 400 was defined in ARES, but here we are using nm instead of Angstroms
// therefore 15000 is a good guess.
    double fit_ngauss_width = cpl_parameter_get_double(
                               cpl_parameterlist_find_const(
                              paramlist,"espda.espda_compu_eqwidth.fit_ngauss_width"));

//    if (fit_ngauss_width < 1000 || fit_ngauss_width > 200000 ) {
    if (fit_ngauss_width < 0.1 || fit_ngauss_width > 2.5 ) {
        return (int)cpl_error_set_message(cpl_func,
               CPL_ERROR_INCOMPATIBLE_INPUT, "-fit_ngauss_width- parameter out of bounds: [0.1,2.5]");
//               CPL_ERROR_INCOMPATIBLE_INPUT, "-fit_ngauss_width- parameter out of bounds: [1000,200000]");
    }

    // change of variable for the fit
    fit_ngauss_width = 1./(2.*fit_ngauss_width/100.*fit_ngauss_width/100.);

    // We need to call a function to take out s/n and compute the required value
    // for now it is fixed as 0.995
    if (cont_rejt < 0.0) {

      proplist = cpl_propertylist_load_regexp(
                    cpl_frame_get_filename(spec_frame), 0, DAS_SNR600_RMS, 0);

      double s_n = 550; //need to get the value from the header
      if (proplist == NULL) {
        cpl_msg_info(cpl_func, "SNR missing from input RSPEC_PRE spectra. Using a s_n of 550. (You may want to adjust cont rejt parameter)\n");
      } else {
        s_n = cpl_propertylist_get_double(proplist, DAS_SNR600_RMS);
        cpl_msg_info(cpl_func, "using SNR from RSPEC_PRE spectra\n");
      }

      cpl_msg_info(cpl_func, "using SNR of: %f \n", s_n);

      if (s_n < 100) {
        cont_rejt = 0.948378 + 0.000639270*s_n - 0.00000241632*s_n*s_n;
      } else {
        if (s_n >= 100 && s_n <  125) cont_rejt = 0.990;
        if (s_n >= 125 && s_n <  150) cont_rejt = 0.991;
        if (s_n >= 150 && s_n <  200) cont_rejt = 0.992;
        if (s_n >= 200 && s_n <  225) cont_rejt = 0.993;
        if (s_n >= 225 && s_n <  250) cont_rejt = 0.994;
        if (s_n >= 250 && s_n <  300) cont_rejt = 0.995;
        if (s_n >= 300 && s_n <  500) cont_rejt = 0.996;
        if (s_n >= 500 && s_n <  800) cont_rejt = 0.997;
        if (s_n >= 800 && s_n < 1500) cont_rejt = 0.998;
        if (s_n >= 1500) cont_rejt = 0.999;
      }
      printf("cont_rejt to be used: %f\n", cont_rejt);
    }





    // Loading the lines to be measured into a cpl table
    linelist_table = cpl_table_load(cpl_frame_get_filename(lines_frame), 1, 0);
    printf("Filename LineList Fits Table: %s\n", cpl_frame_get_filename(lines_frame));

    // Creating new columns for storing local spectrum
    cpl_table_new_column_array(linelist_table, "WAVEL_LOC", CPL_TYPE_DOUBLE, LOC_SPEC_NPIX);
    cpl_table_new_column_array(linelist_table, "FLUX_LOC", CPL_TYPE_DOUBLE, LOC_SPEC_NPIX);
    cpl_table_new_column_array(linelist_table, "FIT_LOC", CPL_TYPE_DOUBLE, LOC_SPEC_NPIX);

//    line = 450.828; // Blended line
//    line = 670.511; // Single line
//    line = 670.000;   // No line
//    line = 602.406; // Blended noisy

    double ll_spec_begin = cpl_table_get(spec_table, "WAVEL",0, error);
    double ll_spec_end = cpl_table_get(spec_table, "WAVEL",cpl_table_get_nrow(spec_table)-1, error);

    printf("Limits of the spectrum: %f, %f \n", ll_spec_begin, ll_spec_end);

    int i, nlines = cpl_table_get_nrow(linelist_table);
//    printf("%d  %d \n", 69, cpl_table_get_nrow(spec_table));
    for (i = 0; i < nlines; i++) {
//    for (i = 0; i < 3; i++) {

    	// NEED TO ADD HERE THE LIMITS OF ll_begin and ll_end (wavel_min, wavel_max)
        line = cpl_table_get_double(linelist_table,"WAVEL",i,NULL);

        if (line < wavel_min || line > wavel_max) {
            cpl_msg_info(cpl_func, "Line: %.5f, outside limits defined by parameters wavel_min and wavel_max [%6.2f, %6.2f]...\n", line, wavel_min, wavel_max);
            continue;
        }

        if (line < ll_spec_begin + space || line > ll_spec_end - space) {
            cpl_msg_info(cpl_func, "Line: %.5f, outside limits of the spectra [%6.2f,%6.2f]...\n", line, ll_spec_begin, ll_spec_end);
            continue;
        }

//        printf("line: %f\n",line);

        cpl_ensure_code(espda_select_chunk(spec_table,
                                            space,
                                            line,
                                            &spec_region)
                        == CPL_ERROR_NONE, cpl_error_get_code());




//    cpl_msg_info(cpl_func, "*** 1111 ***");


        cpl_ensure_code(espda_fit_lcont(spec_region,
                                      cont_rejt,
                                      cont_iter)
                        == CPL_ERROR_NONE, cpl_error_get_code());




//    cpl_msg_info(cpl_func, "*** 2222 ***");


        int xind1, xind2;
        find_left_right_continuum_pos(&xind1, &xind2, spec_region, cont_rejt, line);
    //    printf("Index of continuum: %d , %d \n", xind1, xind2);
        spec_cont_region = cpl_table_extract(spec_region,
                                              xind1,
                                              xind2-xind1);

    //    cpl_plot_column("","","",spec_region,"WAVEL","FLUX");
    //    cpl_plot_column("","","",spec_cont_region,"WAVEL","FLUX");


        cpl_ensure_code(espda_det_line(spec_cont_region,
                                      det_line_thres,
                                      det_line_resol,
                                      det_line_smwidth,
                                      &line_table)
                        == CPL_ERROR_NONE, cpl_error_get_code());

        if (cpl_table_get_nrow(line_table) > 0) {

            cpl_ensure_code(espda_fit_ngauss(spec_cont_region,
                                           line_table,
                                           fit_ngauss_width)
                            == CPL_ERROR_NONE, cpl_error_get_code());


    //        cpl_table_dump(line_table,0,1000,NULL);

            int index_line, n_lines;
            double ew_err = 0.;
            double ew = check_ew(line_table, line, det_line_resol, &index_line, &n_lines, &ew_err);

/*            printf("Line: %.5f, Line Center: %.5f, EW: %.3f +- %.3f, nfit: %d \n",
                    line,
                    cpl_table_get_double(line_table,"WAVEL",index_line,NULL),
                    ew, ew_err,
                    n_lines);
*/

            cpl_msg_info(cpl_func,"*** Line (%d/%d): %.3f nm, Line Found: %.3f nm, EW: %.1f +- %.1f, nfit: %d ***",
                    i,nlines,
                    line,
                    cpl_table_get_double(line_table,"WAVEL",index_line,NULL),
                    ew, ew_err,
                    n_lines);

            if (ew_min < ew) {
                cpl_table_set_double(linelist_table,"PEAK",i,
                                     cpl_table_get_double(line_table,"PEAK",index_line,NULL));
                cpl_table_set_double(linelist_table,"FWHM",i,
                                     cpl_table_get_double(line_table,"FWHM",index_line,NULL));
                cpl_table_set_double(linelist_table,"EW",i,ew);
                cpl_table_set_double(linelist_table,"EW_ERR",i,ew_err);

                // Get arrays for local spectra HERE:
                int ia,ns = cpl_table_get_nrow(spec_region);
                cpl_array *wl = cpl_array_new(LOC_SPEC_NPIX, CPL_TYPE_DOUBLE);
                cpl_array *fl = cpl_array_new(LOC_SPEC_NPIX, CPL_TYPE_DOUBLE);
                cpl_array *fit= cpl_array_new(LOC_SPEC_NPIX, CPL_TYPE_DOUBLE);
                for (ia=0; ia<ns; ia++) {
                  cpl_array_set(wl, ia, cpl_table_get_double(spec_region,"WAVEL",ia,NULL));
                  cpl_array_set(fl, ia, cpl_table_get_double(spec_region,"FLUX" ,ia,NULL));
                }
                for (ia=ns; ia<LOC_SPEC_NPIX; ia++) {
                  cpl_array_set(wl, ia, cpl_array_get_double(wl,ia-1,0)+0.001);
                  cpl_array_set(fl, ia, 1);
                }
                int il, nl = cpl_table_get_nrow(line_table);
                for (ia=0; ia<LOC_SPEC_NPIX; ia++) {
                  double wave = cpl_array_get_double(wl,ia,0);
                  double sumf = 1;
                  for (il = 0; il < nl; il++) {
                    double wc = cpl_table_get_double(line_table,"WAVEL",il,NULL);
                    double peak = cpl_table_get_double(line_table,"PEAK",il,NULL);
                    double lambda = 0.5/(cpl_table_get_double(line_table,"SIGMA",il,NULL)*cpl_table_get_double(line_table,"SIGMA",il,NULL));
                    sumf += -peak * exp ( -lambda * (wave - wc) * (wave - wc) );
                  }
                  cpl_array_set(fit, ia, sumf);
                }

                cpl_table_set_array(linelist_table, "WAVEL_LOC", i, wl);
                cpl_table_set_array(linelist_table, "FLUX_LOC", i, fl);
                cpl_table_set_array(linelist_table, "FIT_LOC", i, fit);
                cpl_array_delete(wl);
                cpl_array_delete(fl);
                cpl_array_delete(fit);

            } else {
                cpl_msg_info(cpl_func, "Line: %.5f nm: Weak Line (< ew_min (%6.2f), not saved in product.", line, ew_min);
            }

        } else {
            cpl_msg_info(cpl_func, "Line: %.5f nm, Not Detected...", line);
            //Maybe it is necessary to do something here... FLAG the line somehow???
        }

//  iteration memory free

    cpl_table_delete(spec_cont_region);
    cpl_table_delete(spec_cont_region_der1);
    cpl_table_delete(spec_cont_region_der2);
    cpl_table_delete(spec_cont_region_der3);
    cpl_table_delete(spec_region);
    cpl_table_delete(line_table);

    }
//    cpl_table_dump(linelist_table,0,1000,NULL);

/*    cpl_ensure_code(cpl_table_save(linelist_table,
                                   NULL,
                                   NULL,
                                   "Test_EW.fits",
                                   CPL_IO_CREATE)
                    == CPL_ERROR_NONE, cpl_error_get_code());
*/

// Saving the product as a fits table
    proplist = cpl_propertylist_new();
    cpl_propertylist_append_string(proplist, CPL_DFS_PRO_CATG, "LINE_EQWIDTH");
    cpl_ensure_code(
        cpl_dfs_save_table(frameset, NULL, paramlist, frameset,
                           NULL, linelist_table, NULL,
                           "espda_compu_eqwidth", proplist, NULL, PACKAGE "/"
                           PACKAGE_VERSION, "LINE_EQWIDTH.fits")
        == CPL_ERROR_NONE, cpl_error_get_code()
    );


// Final Memory Free
    cpl_propertylist_delete (proplist);
    cpl_table_delete(spec_table);
    cpl_table_delete(linelist_table);

    cpl_msg_info(cpl_func, "*** Success!! ***");
    return (int)cpl_error_get_code();

}

/**@}*/
