/* $Id: espda_compu_radvel.c 288160 2024-03-28 08:38:30Z amodigli $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: amodigli $
 * $Date: 2024-03-28 08:38:30 +0000 (Thu, 28 Mar 2024) $
 * $Revision: 288160 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "espda_ccf.h"
#include "espda_check.h"
#include "espda_dfs.h"
#include "espda_frame.h"
#include "espda_param.h"

/*-----------------------------------------------------------------------------
                             Plugin registration
 -----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist * list);

/*-----------------------------------------------------------------------------
                            Private function prototypes
 -----------------------------------------------------------------------------*/

static int espda_compu_radvel_create(cpl_plugin *);
static int espda_compu_radvel_exec(cpl_plugin *);
static int espda_compu_radvel_destroy(cpl_plugin *);
static int espda_compu_radvel(cpl_frameset *, const cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/
static char espda_compu_radvel_short_description[] =
"computes the radial velocity for a stellar spectrum by cross-correlating it with a mask of the specific stellar type.";
static char espda_compu_radvel_description[] =
"This recipe computes the radial velocity for a stellar spectrum by          \n"
"cross-correlating it with a mask of the specific stellar type.              \n"
"\n"
"INPUT FRAMES                                             # EXPECTED     TAGS\n"
"Spectrum in the [extracted pixel,spectral order] space     1       S2D_BLAZE\n"
"Blaze function                                             1           BLAZE\n"
"Mask Look up table                                         1        MASK_LUT\n"
"Line mask tables for specific stellar types               11      MASK_TABLE\n"
"Flux distribution as a function of wavelengths order       1   FLUX_TEMPLATE\n"
"for different stellar types                                                 \n"
"\n"
"OUTPUT FRAMES                                            # EXPECTED     TAGS\n"
"Cross-correlation function                                 1             CCF\n"
;

/*-----------------------------------------------------------------------------
                                Function code
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok, 1 otherwise
  @note     Only this function is exported

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist *list)
{
    cpl_recipe *recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin *plugin = &recipe->interface;

    if (cpl_plugin_init(plugin, CPL_PLUGIN_API, ESPDA_BINARY_VERSION,
                        CPL_PLUGIN_TYPE_RECIPE, "espda_compu_radvel",
						espda_compu_radvel_short_description,
                        espda_compu_radvel_description, "Danuta Sosnowska",
                        PACKAGE_BUGREPORT, espda_dfs_license(),
                        espda_compu_radvel_create, espda_compu_radvel_exec,
                        espda_compu_radvel_destroy)) {    
        cpl_msg_error(cpl_func, "Plugin initialization failed");
        (void)cpl_error_set_where(cpl_func);                          
        return 1;                                               
    }                                                    

    if (cpl_pluginlist_append(list, plugin)) {                 
        cpl_msg_error(cpl_func, "Error adding plugin to list");
        (void)cpl_error_set_where(cpl_func);                         
        return 1;                                              
    }                                                          
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_radvel_create(cpl_plugin *plugin)
{
    cpl_recipe    *recipe;                                               
    cpl_parameter *param;
                                                                     
    /* Do not create the recipe if an error code is already set */     
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                      
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());      
        return (int)cpl_error_get_code();                              
    }                                                                  
                                                                       
    if (plugin == NULL) {                                              
        cpl_msg_error(cpl_func, "Null plugin");                        
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                 
    }                                                                  
                                                                       
    /* Verify plugin type */                                           
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {       
        cpl_msg_error(cpl_func, "Plugin is not a recipe");             
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);              
    }                                                                  
                                                                       
    /* Get the recipe */                                               
    recipe = (cpl_recipe *)plugin;                                     
                                                                       
    /* Create the parameters list in the cpl_recipe object */          
    recipe->parameters = cpl_parameterlist_new();                      
    if (recipe->parameters == NULL) {                                  
        cpl_msg_error(cpl_func, "Parameter list allocation failed");   
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);             
    }                                                                  

    /* Fill the parameters list */

/*    
 * Default values for all parameters are set to INV_DOUBLE = -9999.0, because no
 * other default has been determined yet.
 */
    param = cpl_parameter_new_value("espda.espda_compu_radvel.rv_center",
        CPL_TYPE_DOUBLE, "Center of the radial velocity window [km/s]", 
        "espda.espda_compu_radvel", 9999.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "rv_center");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_compu_radvel.rv_step",
        CPL_TYPE_DOUBLE, "Radial velocity step [km/s]",
        "espda.espda_compu_radvel", 0.5);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "rv_step");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_compu_radvel.rv_range",
        CPL_TYPE_DOUBLE, "Width of the radial velocity range [km/s]", 
        "espda.espda_compu_radvel", 20.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "rv_range");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);
    
    param = cpl_parameter_new_value("espda.espda_compu_radvel.mask_table_id",
                                    CPL_TYPE_STRING, "Correlation mask type [F9, G2, M2, ...]",
                                    "espda.espda_compu_radvel", "XX");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "mask_table_id");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);
    
    param = cpl_parameter_new_value("espda.espda_compu_radvel.mask_width",
        CPL_TYPE_DOUBLE, "Width of mask holes [km/s]", 
        "espda.espda_compu_radvel", 0.5);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "mask_width");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_compu_radvel.flux_correction_type",
                                    CPL_TYPE_STRING, "Flux correction type [AUTO/NONE/sp_type]",
                                    "espda.espda_compu_radvel", "AUTO");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "flux_correction_type");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_radvel_exec(cpl_plugin *plugin) {
    
    int recipe_status;
    cpl_recipe *recipe;
    cpl_errorstate initial_errorstate = cpl_errorstate_get();
                                                                           
    /* Return immediately if an error code is already set */               
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                          
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",    
                      cpl_func, __LINE__, cpl_error_get_where());          
        return (int)cpl_error_get_code();                                  
    }                                                                      
                                                                           
    if (plugin == NULL) {                                                  
        cpl_msg_error(cpl_func, "Null plugin");                            
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    /* Verify plugin type */                                               
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {           
        cpl_msg_error(cpl_func, "Plugin is not a recipe");                 
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);                  
    }                                                                      
                                                                           
    /* Get the recipe */                                                   
    recipe = (cpl_recipe *)plugin;                                         
                                                                           
    /* Verify parameter and frame lists */                                 
    if (recipe->parameters == NULL) {                                      
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      

    if (recipe->frames == NULL) {                                          
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");     
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    /* Invoke the recipe */                                                
    recipe_status = espda_compu_radvel(recipe->frames, recipe->parameters);
                                                                           
    /* Ensure DFS-compliance of the products */                            
    if (cpl_dfs_update_product_header(recipe->frames)) {                   
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();
    }                                                                      
                                                                           
    if (!cpl_errorstate_is_equal(initial_errorstate)) {                    
        /* Dump the error history since recipe execution start.            
           At this point the recipe cannot recover from the error */       
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);          
    }                                                                      
                                                                           
    return recipe_status;                                                  
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_radvel_destroy(cpl_plugin *plugin)
{
    cpl_recipe * recipe;                                          
                                                                  
    if (plugin == NULL) {                                         
        cpl_msg_error(cpl_func, "Null plugin");                   
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);            
    }                                                             
                                                                  
    /* Verify plugin type */                                      
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {  
        cpl_msg_error(cpl_func, "Plugin is not a recipe");        
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);         
    }                                                             
                                                                  
    /* Get the recipe */                                          
    recipe = (cpl_recipe *)plugin;                                
                                                                  
    cpl_parameterlist_delete(recipe->parameters);             
                                                                  
    return 0;                                                    
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist    the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_radvel(cpl_frameset *frameset,
                              const cpl_parameterlist *parlist) {

/*
 * Input is checked and loaded. 
 */
   
    cpl_msg_set_level(CPL_MSG_DEBUG);
    /* check for invalid input */
    if(espda_files_dont_exist(frameset) != CPL_ERROR_NONE) {
    	return CPL_ERROR_BAD_FILE_FORMAT;
    }

    espda_msg("Starting radvel computation");
    
    // Load and check frames
    espda_dfs_ensure(espda_dfs_groups(frameset));
    
/*
 * To load the frames, please uncomment the following lines. Instructions:
 * 
 *  1. 1ST_TAG, 2ND_TAG, 3RD_TAG are the tags of your input frames; you can add
 *     as many tags as you want.
 *  2. 1ST_TAG must refer to a (set of) required frame(s); it is configured with
 *     cpl_recipeconfig_set_tag(), which sets the minimum/maximum number of 
 *     frames for the given tag.
 *  2. Other tags are configured with cpl_recipeconfig_set_input(), which also 
 *     sets the minimum/maximum number of frames for the given tags.
 *  3. Frameset is checked for frame names, frame tags, and frame numbers.
 *  4. Frames are loaded by espda_frame_load() (from espda_frame.c) under the 
 *     assumption that they are tables; you should write a similar function for
 *     loading images.
 *
    char *cfg_tag = 1ST_TAG;
    cpl_recipeconfig_set_tag  (cfg, cfg_tag, 1, 1);  
    cpl_recipeconfig_set_input(cfg, cfg_tag, 2ND_TAG, 0, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, 3RD_TAG,  0, 1);
    espda_dfs_ensure(espda_check_name(frameset, cfg, cfg_tag));
    espda_dfs_ensure(espda_check_tag (frameset, cfg, cfg_tag));
    espda_dfs_ensure(espda_check_num (frameset, cfg, cfg_tag));
    cpl_table *1st_frame = espda_frame_load(frameset, cfg, cfg_tag, -1);
    cpl_table *2nd_frame = espda_frame_load(frameset, cfg, cfg_tag,  0);
    cpl_table *3rd_frame = espda_frame_load(frameset, cfg, cfg_tag,  1);
    cpl_recipeconfig_delete(cfg);
 */
    
    cpl_frameset *S2D_frameset = cpl_frameset_new(); 	
    cpl_frameset *MASK_LUT_frameset = cpl_frameset_new();
    cpl_frameset *MASK_TABLE_frameset = cpl_frameset_new();
    cpl_frameset *BLAZE_frameset = cpl_frameset_new();
    cpl_frameset *FLUX_frameset = cpl_frameset_new();
    
    cpl_image *flux = NULL;
    cpl_image *flux_err = NULL;
    cpl_image *flux_qual = NULL;
    cpl_image *ll = NULL;
    cpl_image *dll = NULL;
    cpl_table *mask_lut_table = NULL;
    cpl_table *mask_table = NULL;
    cpl_table *flux_table = NULL;
    
    cpl_propertylist *keywords = NULL;
    
    const char *s2d_name;
    const char *mask_lut_name;
    const char *mask_table_name;
    const char *blaze_name;
    const char *flux_name;
    
    cpl_error_code my_error = CPL_ERROR_NONE;
    
    /* Loading the S2D frame */
    my_error = espda_frame_extract_by_tag(frameset, "S2D_BLAZE",
                                          S2D_frameset);
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "S2D frame extraction failed");
    espda_ensure(cpl_frameset_get_size(S2D_frameset) < 1, my_error,
                 "No S2D frame with tag = S2D_BLAZE, exiting");
    
    s2d_name = cpl_frame_get_filename(cpl_frameset_get_position(S2D_frameset, 0));
    
    espda_msg("s2d_name: %s",s2d_name);
    
    flux = cpl_image_load(s2d_name,CPL_TYPE_DOUBLE,0,1);
    flux_err = cpl_image_load(s2d_name,CPL_TYPE_DOUBLE,0,2);
    flux_qual = cpl_image_load(s2d_name,CPL_TYPE_INT,0,3);
    ll = cpl_image_load(s2d_name,CPL_TYPE_DOUBLE,0,5);
    dll = cpl_image_load(s2d_name,CPL_TYPE_DOUBLE,0,7);
    
    int nx = cpl_image_get_size_x(flux);
    int ny = cpl_image_get_size_y(flux);
    
    /* Getting keywords for S2D frame */
    keywords = cpl_propertylist_load(s2d_name, 0);
    
    /* Loading the BLAZE frame */
    my_error = espda_frame_extract_by_tag(frameset, "BLAZE",
                                          BLAZE_frameset);
    
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "BLAZE frame extraction failed");
    espda_ensure(cpl_frameset_get_size(BLAZE_frameset) < 1, my_error,
                 "No BLAZE frame with tag = BLAZE, exiting");
    
    blaze_name = cpl_frame_get_filename(cpl_frameset_get_position(BLAZE_frameset, 0));
    espda_msg("blaze_name: %s",blaze_name);
    cpl_image *blaze = cpl_image_load(blaze_name,CPL_TYPE_DOUBLE,0,1);
    
    /* Loading the MASK LookUp table */
    my_error = espda_frame_extract_by_tag(frameset, "MASK_LUT",
                                          MASK_LUT_frameset);
    
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "MASK LUT frame extraction failed");
    espda_ensure(cpl_frameset_get_size(MASK_LUT_frameset) < 1, my_error,
                 "No mask look up frame with tag = MASK_LUT, exiting");

    mask_lut_name = cpl_frame_get_filename(cpl_frameset_get_position(MASK_LUT_frameset, 0));
    espda_msg("mask_lut_name: %s", mask_lut_name);
    mask_lut_table = cpl_table_load(mask_lut_name,1,0);
    
    // Load and check parameters

/*
 * Please adjust the parameter range as needed. Beware that the default value 
 * for all parameters is currently INV_DOUBLE=-9999.0 (see above)
 */ 

    espda_param_double("rv_center", rv_center,
                       cpl_parameterlist_get_first_const(parlist),
                       -10000.0, 10000.0);
    espda_param_double("rv_step", rv_step,
                       cpl_parameterlist_get_next_const(parlist),
                       -5.0, 5.0);
    espda_param_double("rv_range", rv_range,
                       cpl_parameterlist_get_next_const(parlist),
                       1, 1000);
    espda_param_string("mask_table_id", mask_table_id,
                       cpl_parameterlist_get_next_const(parlist));
    espda_param_double("mask_width", mask_width,
                       cpl_parameterlist_get_next_const(parlist),
                       -10.0, 10.0);
    espda_param_string("flux_correction_type", flux_correction_type,
                       cpl_parameterlist_get_next_const(parlist));
    
    espda_msg("Params loaded");

/*
 * Data is processed. Please put all your functions here. I put just 
 * espda_compute_CCF(), with dummy input.
 *
 */

    cpl_image *CCF_flux_RE = NULL, *CCF_error_RE = NULL, *CCF_quality_RE = NULL;
    double BERV, BERVMAX, RADVEL;
    const char *INSTRUME;

    INSTRUME = cpl_propertylist_get_string(keywords,"INSTRUME");
    BERV = cpl_propertylist_get_double(keywords,"ESO QC BERV");
    BERVMAX = cpl_propertylist_get_double(keywords,"ESO QC BERVMAX");
    
    espda_msg("BERV = %f\tBERVMAX = %f",
              BERV, BERVMAX);
    
    /* Getting the keyword containing the spectral type */
    char spec_type[24];
    const char *spt_kw;
    
    if (strcmp(INSTRUME, "HARPS") == 0) {
        
        char *separator;
        char to_split[64];
        
        spt_kw = cpl_propertylist_get_string(keywords, "ESO DPR TYPE");
        my_error = cpl_error_get_code();
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "No ESO DPR TYPE kw found, exiting");
        
        strcpy(to_split, spt_kw);
        separator = strtok(to_split,",");
        while(separator != NULL) {
            strcpy(spec_type, separator);
            separator = strtok(NULL,",");
        }
        espda_msg("SPECTRAL TYPE = %s", spec_type);
        
    } else {
        if (strcmp(INSTRUME, "NIRPS") == 0) {
            /*espda_msg("ESO OBS TARG NAME = %s", cpl_propertylist_get_string(keywords, "ESO OBS TARG NAME"));
            if (strcmp(cpl_propertylist_get_string(keywords, "ESO OBS TARG NAME"), "SUN") == 0) {
                espda_msg("Exception for the SUN, reading keyword ESO TEL TARG SPTYPE = %s", cpl_propertylist_get_string(keywords, "ESO TEL TARG SPTYPE"));
                spt_kw = cpl_propertylist_get_string(keywords, "ESO TEL TARG SPTYPE");
                my_error = cpl_error_get_code();
                espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                             "No ESO TEL TARG SPTYPE kw found, exiting.");
            } else {*/
            espda_msg("Reading keyword ESO OCS TARG SPTYPE = %s", cpl_propertylist_get_string(keywords, "ESO OCS TARG SPTYPE"));
            spt_kw = cpl_propertylist_get_string(keywords, "ESO OCS TARG SPTYPE");
            my_error = cpl_error_get_code();
            espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                          "No ESO OCS TARG SPTYPE kw found, exiting.");
            //}
        } else {
	    spt_kw = cpl_propertylist_get_string(keywords, "ESO OCS OBJ SP TYPE");
            my_error = cpl_error_get_code();
            espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                         "No ESO OCS OBJ SP TYPE kw found, exiting.");
	}

        strcpy(spec_type,spt_kw);
        espda_msg("SPECTRAL TYPE = %s", spec_type);
    }
    
    /* Defining the default spectral type in case of NULL or empty */
    
    if(strcmp(spec_type,"NULL") == 0 || /* spec_type NULL */
       strcmp(spec_type,"")     == 0 /* spec_type blank */
       ) {
        
        espda_msg(ANSI_COLOR_RED"Warning: spectral type: %s, not valid."ANSI_COLOR_RESET,
                  spec_type);
        strcpy(spec_type, "G2");
        espda_msg(ANSI_COLOR_RED"Setting default spectral type"ANSI_COLOR_RESET);
        espda_msg(ANSI_COLOR_RED"SPECTRAL TYPE = %s"ANSI_COLOR_RESET,spec_type);
    }
    
    char mask_table_id_from_FITS[5];
    /* ASE: for mask_table_id, make sure to take the 2 first chars */
    memset(mask_table_id_from_FITS, '\0', sizeof(mask_table_id_from_FITS));
    strncpy(mask_table_id_from_FITS, spec_type, 2);
    
/*
 * Overload paramaters.
 */

    //double rv_center, rv_step, rv_range, mask_width; 	

    const cpl_parameter * par_rv_center =
        cpl_parameterlist_find_const(parlist,"espda.espda_compu_radvel.rv_center");
    int rv_center_set = cpl_parameter_get_default_flag(par_rv_center);

    const cpl_parameter * par_rv_step =
        cpl_parameterlist_find_const(parlist,"espda.espda_compu_radvel.rv_step");
    int rv_step_set = cpl_parameter_get_default_flag(par_rv_step);

    const cpl_parameter * par_rv_range =
        cpl_parameterlist_find_const(parlist,"espda.espda_compu_radvel.rv_range");
    int rv_range_set = cpl_parameter_get_default_flag(par_rv_range);

    const cpl_parameter * par_mask_table_id =
        cpl_parameterlist_find_const(parlist, "espda.espda_compu_radvel.mask_table_id");
    int mask_table_id_set = cpl_parameter_get_default_flag(par_mask_table_id);
    int mask_table_id_has_changed = 0;
    if (strcmp(cpl_parameter_get_string(par_mask_table_id),
               cpl_parameter_get_default_string(par_mask_table_id)) != 0) {
        mask_table_id_has_changed = 1;
    }

    const cpl_parameter * par_mask_width =
         cpl_parameterlist_find_const(parlist,"espda.espda_compu_radvel.mask_width");
    int mask_width_set = cpl_parameter_get_default_flag(par_mask_width);

    const cpl_parameter * par_flux_correction_type =
        cpl_parameterlist_find_const(parlist, "espda.espda_compu_radvel.flux_correction_type");
    int flux_correction_type_set = cpl_parameter_get_default_flag(par_flux_correction_type);
    
    if (rv_center_set == 1)  {
	    espda_param_double("rv_center",rv_center,
                           cpl_parameterlist_find_const(parlist, "espda.espda_compu_radvel.rv_center"),
                           -10000.0, 10000.0);
    }

    if (rv_step_set == 1) {
        espda_param_double("rv_step",rv_step,
                           cpl_parameterlist_find_const(parlist, "espda.espda_compu_radvel.rv_step"),
                           -5.0, 5.0);
    }

    if (rv_range_set == 1) {
        espda_param_double("rv_range",rv_range,
                           cpl_parameterlist_find_const(parlist, "espda.espda_compu_radvel.rv_range"),
                           1, 1000);
    }

    if (mask_table_id_set == 1 && mask_table_id_has_changed == 1) {
        espda_param_string("mask_table_id", mask_table_id,
                           cpl_parameterlist_find_const(parlist, "espda.espda_compu_radvel.mask_table_id"));
    }
    espda_msg("User given mask_table_id = %s",
              cpl_parameter_get_string(par_mask_table_id));
    
    if (mask_width_set == 1) {
        espda_param_double("mask_width", mask_width,
                           cpl_parameterlist_find_const(parlist, "espda.espda_compu_radvel.mask_width"),
                           -10.0, 10.0);
    }

    if (flux_correction_type_set == 1) {
        espda_param_string("flux_correction_type", flux_correction_type,
                           cpl_parameterlist_find_const(parlist, "espda.espda_compu_radvel.flux_correction_type"));
    }
    
   /* Picking the right RADVEL */

    if ((strcmp(INSTRUME,"HARPS") == 0) || (strcmp(INSTRUME,"NIRPS") == 0)) {
        if (rv_center != 9999) {
            RADVEL = rv_center;
        } else {
            RADVEL = cpl_propertylist_get_double(keywords,"ESO TEL TARG RADVEL");
        }
    } else if (strcmp(INSTRUME,"ESPRESSO") == 0) {
        if (rv_center != 9999) {
            RADVEL = rv_center;
        } else {
            RADVEL = cpl_propertylist_get_double(keywords,"ESO OCS OBJ RV");
        }
    } else {
        cpl_msg_info(cpl_func, "Wrong instrument");
        exit(EXIT_FAILURE);
    }
    
    /* Extract flux template file frame */
    int flux_template_nb;
    char flux_corr_type[10];
    sprintf(flux_corr_type, "NONE");
    
    if ((strcmp(flux_correction_type, "NONE") != 0) &&
        (strcmp(flux_correction_type, "none") != 0)) {
        
        my_error = espda_frame_extract_by_tag(frameset,
                                              "FLUX_TEMPLATE",
                                              FLUX_frameset);
        espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                     "Flux template frame extraction failed");
        
        int flux_template_rows;
        int flux_template_cols;
        
        flux_name = cpl_frame_get_filename(cpl_frameset_get_position(FLUX_frameset,0));
        espda_msg("Flux template : %s", flux_name);
        
        flux_table = cpl_table_load(flux_name, 1, 0);
        
        espda_ensure(cpl_error_get_code() != CPL_ERROR_NONE,
                     cpl_error_get_code(),
                     "Can't load FLUX_TEMPLATE table, exiting.");
        
        flux_template_rows = cpl_table_get_nrow(flux_table);
        flux_template_cols = cpl_table_get_ncol(flux_table);
        
        espda_msg("Flux template rows number =%d",
                  flux_template_rows);
        espda_msg("Flux template cols number =%d",
                  flux_template_cols);
        
        if ((strcmp(flux_correction_type, "AUTO") == 0) ||
            (strcmp(flux_correction_type, "auto") == 0)) {
            memset(flux_corr_type, '\0', sizeof(flux_corr_type));
            strncpy(flux_corr_type, spec_type, 2);
        } else {
            sprintf(flux_corr_type, "%s", flux_correction_type);
        }
        
    } else {
        espda_msg("User set no flux correction (flux_correction_type parameter is NONE)");
    }
    
    
    /* Extract mask table */
    
    const char *default_mask_id = "G2";
    const char *mask_id;
    if (mask_table_id_set == 0 || mask_table_id_has_changed == 0) {
        mask_id = espda_get_mask_name(mask_table_id_from_FITS, mask_lut_table);
        espda_msg("Got mask name for %s: %s",
                  mask_table_id_from_FITS, mask_id);
        
        if (strcmp(mask_id, "NOT_FOUND") == 0) {
            espda_msg("Spectral type not found. Default mask is %s",
                      default_mask_id);
            mask_id = default_mask_id;
        }
    } else {
        mask_id = mask_table_id;
    }
    
    espda_msg( "Mask file name to be selected : %s", mask_id);
    
    /* Extract mask file frame */
    int mask_table_nb;
    
    my_error = espda_frame_extract_by_tag(frameset,
                                          "MASK_TABLE",
                                          MASK_TABLE_frameset);
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "MASK TABLE frame extraction failed");
    espda_ensure(cpl_frameset_get_size(MASK_TABLE_frameset) < 1, my_error,
                 "No mask frame with tag = MASK_TABLE, exiting");
    
    mask_table_nb = cpl_frameset_get_size(MASK_TABLE_frameset);
    mask_table = (cpl_table *)cpl_malloc(sizeof(cpl_table *));
    
    int mask_table_rows;
    int mask_table_cols;
    int imask = -1, idefault = -1;
    
    for (int i = 0 ; i < mask_table_nb ; i++) {
        
        mask_table_name =
        cpl_frame_get_filename(cpl_frameset_get_position(MASK_TABLE_frameset,i));
        
        if(strstr(mask_table_name,mask_id) != NULL) {
            imask = i;
            //espda_msg("Mask found: %s", mask_table_name);
        }
        if(strstr(mask_table_name,default_mask_id) != NULL) {
            idefault = i;
            //espda_msg("Default mask found: %s", mask_table_name);
        }
    }
    
    if ((imask == -1) && (idefault != -1)) {
        espda_msg("Mask table %s not found, using default mask G2", mask_id);
        imask = idefault;
        mask_id = default_mask_id;
    }
    if ((imask == -1) && (idefault == -1)) {
        espda_msg("Mask table %s nor default mask table not found, exiting", mask_id);
        return(CPL_ERROR_INCOMPATIBLE_INPUT);
    }
    
    mask_table_name =
    cpl_frame_get_filename(cpl_frameset_get_position(MASK_TABLE_frameset,imask));
    espda_msg("Mask table selected : %s", mask_table_name);
    
    mask_table = cpl_table_load(mask_table_name,1,0);
    my_error = cpl_error_get_code();
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "Correlation mask loading failed: %s - check the format",
                 cpl_error_get_message_default(my_error));
    
    mask_table_rows = cpl_table_get_nrow(mask_table);
    mask_table_cols = cpl_table_get_ncol(mask_table);
    
    espda_msg("Mask table rows number =%d", mask_table_rows);
    espda_msg("Mask table cols number =%d", mask_table_cols);
    
    espda_msg(ANSI_COLOR_GREEN"rv_center = %f"ANSI_COLOR_RESET, RADVEL);
    espda_msg(ANSI_COLOR_GREEN"rv_step = %f"ANSI_COLOR_RESET, rv_step);
    espda_msg(ANSI_COLOR_GREEN"rv_range = %f"ANSI_COLOR_RESET, rv_range);
    espda_msg(ANSI_COLOR_GREEN"mask_width = %f"ANSI_COLOR_RESET, mask_width);
    espda_msg(ANSI_COLOR_GREEN"flux_correction = %s"ANSI_COLOR_RESET,
              flux_correction_type);

    /* Generating RV table, the rv_center, comes as keyword as RADVEL */
    cpl_array *RV_table = NULL;
    my_error = espda_compute_rv_table(rv_range,
                                      rv_step,
                                      RADVEL,
                                      &RV_table);
    
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "espda_compute_rv_table failed: %s",
                 cpl_error_get_message_default(my_error));
    
    int *cosmics_nb = (int *)cpl_calloc(ny, sizeof(int));
    cpl_image *flux_no_cosmics = cpl_image_new(nx, ny, CPL_TYPE_DOUBLE);
    
    espda_msg("Cosmics correction on s2d ...");
    my_error = espda_remove_cosmics_via_histogram(flux,
                                                  flux_err,
                                                  flux_no_cosmics,
                                                  cosmics_nb);
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "espda_remove_cosmics_via_histogram failed: %s",
                 cpl_error_get_message_default(my_error));
    
    espda_msg("Starting flux correction ...");
    
    cpl_image *flux_corr = cpl_image_new(nx, ny, CPL_TYPE_DOUBLE);
    double *flux_correction = (double *)cpl_calloc(ny, sizeof(double));
    
    my_error = espda_correct_flux(flux_no_cosmics,
                                  ll,
                                  flux_table,
                                  mask_table,
                                  flux_corr_type,
                                  flux_corr,
                                  flux_correction);
    
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "espda_correct_flux failed: %s",
                 cpl_error_get_message_default(my_error));
    
    int ny_s2d = cpl_image_get_size_y(flux);
    int nx_ccf = cpl_array_get_size(RV_table);

    CCF_flux_RE = cpl_image_new(nx_ccf,ny_s2d+1,CPL_TYPE_DOUBLE);
    CCF_error_RE = cpl_image_new(nx_ccf,ny_s2d+1,CPL_TYPE_DOUBLE);
    CCF_quality_RE = cpl_image_new(nx_ccf,ny_s2d+1,CPL_TYPE_INT);

    /* Shifting ll and dll to berv factor */
    /* The ll and dll from S2D are already shifted */
    //double berv_factor = (1.+1.55e-8)*(1.+BERV/299792.458);
    //cpl_image *dll_shifted = cpl_image_multiply_scalar_create(dll,berv_factor);

    /* Calculating CCF */	

    espda_msg("Computing CCF...");
    my_error = espda_compute_CCF(ll,
                                 dll,
                                 flux_corr,
                                 flux_err,
                                 blaze,
                                 flux_qual,
                                 RV_table,
                                 mask_table,
                                 mask_width,
                                 BERV,
                                 BERVMAX,
                                 &CCF_flux_RE,
                                 &CCF_error_RE,
                                 &CCF_quality_RE);

    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "espda_compute_CCF failed: %s",
                 cpl_error_get_message_default(my_error));
    espda_msg("Compute CCF finished !");
    
    /* Gauss fitting result structure */
    espda_ngauss_result *g_res = (espda_ngauss_result *)
                                    cpl_malloc(sizeof(espda_ngauss_result));
    
    double asym, asym_error, bispan, bispan_error;
    cpl_table *residuals = NULL;

    my_error = espda_fit_ccf(RV_table,
                             CCF_flux_RE,
                             CCF_error_RE,
                             CCF_quality_RE,
                             &asym,
                             &asym_error,
                             &bispan,
                             &bispan_error,
                             &residuals,
                             g_res);
    
    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "espdr_fit_ccf failed: %s",
                 cpl_error_get_message_default(my_error));
    
/*
 * Output is exported. QC parameters are saved. 
 */

    espda_msg("Saving the CCF with computed QC KWs...");
    
    my_error = espda_compu_radvel_QC(RV_table,
                                     rv_step,
                                     ny,
                                     g_res,
                                     mask_id,
                                     cosmics_nb,
                                     flux_correction,
                                     asym,
                                     asym_error,
                                     bispan,
                                     bispan_error,
                                     &keywords);

    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "espda_compu_radvel_QC failed: %s",
                 cpl_error_get_message_default(my_error));
    
    espda_msg("Quality check finished !");
    
/*
 * The following lines may be replaced by your own image-exporting function.  
 * Here I simply save the output of espda_compute_CCF() as separate frames. 
 * Maybe you already defined names and tags in some linked header. 
 */ 
    
    // Save frames
    espda_msg("Saving products...");
    
    char ccf_name[64];
    sprintf(ccf_name,"%s_CCF.fits",INSTRUME);		

    cpl_image **images_to_save;

    images_to_save = (cpl_image**)cpl_malloc(3*sizeof(cpl_image*));

    images_to_save[0] = cpl_image_duplicate(CCF_flux_RE);
    images_to_save[1] = cpl_image_duplicate(CCF_error_RE);
    images_to_save[2] = cpl_image_duplicate(CCF_quality_RE);

    espda_msg("getting PRO CATG: %s",cpl_propertylist_get_string(keywords, CPL_DFS_PRO_CATG));

    // Modify PRO CATG for recorelated CCF, S2D_BLAZE* -> CCF_*
    const char *procatg = cpl_propertylist_get_string(keywords, CPL_DFS_PRO_CATG);
    char *p = strstr(procatg, "S2D_BLAZE");
    if (p != NULL) {
        memcpy(p, "CCF", 3);
        memmove(p + 3, p + 9, strlen(p + 9) + 1); 
    }
    espda_msg("Modified PRO CATG for recorelated CCF = %s\n", procatg);

    cpl_propertylist_update_string(keywords, CPL_DFS_PRO_CATG, procatg);
    cpl_propertylist_update_string(keywords, CPL_DFS_PRO_TYPE, "ANALYZED");

    my_error = espda_dfs_save_data_err_qual(frameset, parlist, frameset,
                                            "espda_compu_radvel",
                                            keywords, ccf_name,
                                            images_to_save);
    espda_msg("CCF saved under: %s", ccf_name);
    
    /* Cleaning memory */
    espda_msg("Cleaning memory");
    cpl_frameset_delete(S2D_frameset);
    cpl_frameset_delete(MASK_LUT_frameset);
    cpl_frameset_delete(MASK_TABLE_frameset);
    cpl_frameset_delete(BLAZE_frameset);
    cpl_frameset_delete(FLUX_frameset);
    cpl_table_delete(mask_lut_table);
    cpl_table_delete(mask_table);
    cpl_table_delete(flux_table);
    cpl_table_delete(residuals);
    cpl_array_delete(RV_table);
    cpl_propertylist_delete(keywords);
    cpl_image_delete(flux_no_cosmics);
    cpl_image_delete(CCF_flux_RE);
    cpl_image_delete(CCF_error_RE);	
    cpl_image_delete(CCF_quality_RE);	
    cpl_free(cosmics_nb);
    cpl_free(g_res->k);
    cpl_free(g_res->sig_k);
    cpl_free(g_res->x0);
    cpl_free(g_res->sig_x0);
    cpl_free(g_res->fwhm);
    cpl_free(g_res->sig_fwhm);
    cpl_free(g_res);
    for (int i = 0; i < 3; i++) {
    	cpl_image_delete(images_to_save[i]);
    }
    cpl_free(images_to_save);
    
    espda_msg("Radvel computation finished correctly");
    
    return (int)cpl_error_get_code();
}

/**@}*/
