/* $Id: espda_synth_spec.c 234673 2019-01-22 17:15:10 jonaydas $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: jonaydas $
 * $Date: 2019-01-16 17:15:10 +0000 (wed, 16 jan 2019) $
 * $Revision: 234673 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#include <string.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include "cpl_plot.h"

#include "espda_interp_synth.h"
#include "espda_convol_synth.h"

#include "espda_dfs.h"
#include "espda_frame.h"
#include "espda_spec.h"
#include "espda_param.h"
#include "espda_struct.h"
#include "espda_syst.h"

/*-----------------------------------------------------------------------------
                             Plugin registration
 -----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist * list);

/*-----------------------------------------------------------------------------
                            Private function prototypes
 -----------------------------------------------------------------------------*/

static int espda_synth_spec_create(cpl_plugin *);
static int espda_synth_spec_exec(cpl_plugin *);
static int espda_synth_spec_destroy(cpl_plugin *);
static int espda_synth_spec(cpl_frameset *, const cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char espda_synth_spec_description[] = 
"This recipe interpolates a grid of synthetic spectra loaded from a CPL FITS Table \n" 
"given as input frame. The input variables are the stellar parameters, metallicity \n" 
"and broadening parameters (instrumental broadening, projected rotational velocity \n"
"(vsini), and macroturbulence) of the target star. \n"
"The recipe performs a trilinear interpolation of synthetic \n"
"spectra for effective temperature, surface gravity and metallicity of the star, and \n"
"convolves the resulting synthetic spectrum with different line profiles using the \n" 
"broadening parameters. This synthetic spectrum is finally interpolated in the \n" 
"wavelength array of observed spectrum. The output of the recipe is a CPL FITS Table \n" 
"with to an interpolated synthetic spectrum.\n"
"\n"
"INPUT FRAMES                          FORMAT  # EXPECTED  TAGS          "
"NOTES: Relevant Table Columns:\n" 
"normalized merged rebinned spectrum   SPEC    1           NMERG_SPEC    "
"WAVEL\n"
"grid of synthetic spectra             SYNTH   1           SYNTH_SPEC    "
"WAVEL, FLUX\n"
"\n"
"OUTPUT FRAMES                         FORMAT  # PRODUCED  TAGS          "
"NOTES: Relevant Table Columns:\n"
"normalized merged rebinned spectrum\n" 
"and interpolated synthetic spectrum   SPEC    1           SMERG_SPEC    "
"WAVEL, FLUX, FLUXERR, SYNTH\n"
;

/*-----------------------------------------------------------------------------
                                Function code
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok, 1 otherwise
  @note     Only this function is exported

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist *list)
{
    cpl_recipe *recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin *plugin = &recipe->interface;

    if (cpl_plugin_init(plugin, CPL_PLUGIN_API, ESPDA_BINARY_VERSION,
                        CPL_PLUGIN_TYPE_RECIPE, "espda_synth_spec", 
                        "Interpolate synthetic spectra for a set of "
                        "stellar parameters", 
                        espda_synth_spec_description, 
                        "Jonay Gonzalez Hernandez",
                        PACKAGE_BUGREPORT, espda_dfs_license(),
                        espda_synth_spec_create, espda_synth_spec_exec,
                        espda_synth_spec_destroy)) {    
        cpl_msg_error(cpl_func, "Plugin initialization failed");
        (void)cpl_error_set_where(cpl_func);                          
        return 1;                                               
    }                                                    

    if (cpl_pluginlist_append(list, plugin)) {                 
        cpl_msg_error(cpl_func, "Error adding plugin to list");
        (void)cpl_error_set_where(cpl_func);                         
        return 1;                                              
    }                                                          
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int espda_synth_spec_create(cpl_plugin *plugin)
{
    cpl_recipe    *recipe;                                               
    cpl_parameter *param;
                                                                     
    /* Do not create the recipe if an error code is already set */     
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                      
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());      
        return (int)cpl_error_get_code();                              
    }                                                                  
                                                                       
    if (plugin == NULL) {                                              
        cpl_msg_error(cpl_func, "Null plugin");                        
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                 
    }                                                                  
                                                                       
    /* Verify plugin type */                                           
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {       
        cpl_msg_error(cpl_func, "Plugin is not a recipe");             
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);              
    }                                                                  
                                                                       
    /* Get the recipe */                                               
    recipe = (cpl_recipe *)plugin;                                     
                                                                       
    /* Create the parameters list in the cpl_recipe object */          
    recipe->parameters = cpl_parameterlist_new();                      
    if (recipe->parameters == NULL) {                                  
        cpl_msg_error(cpl_func, "Parameter list allocation failed");   
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);             
    }                                                                  

    /* Fill the parameters list */
    /* Parameters of espda_interp_synth */
    /* Parameter synth_teff */
    param = cpl_parameter_new_value("espda.espda_synth_spec.synth_teff",
                                    CPL_TYPE_DOUBLE, "Stellar effective "
                                    "temperature [K]: [3500., 7000]",
                                    "espda.espda_synth_spec", 5777.);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "synth-teff");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* Parameter synth_logg */
    param = cpl_parameter_new_value("espda.espda_synth_spec.synth_logg",
                                    CPL_TYPE_DOUBLE, "Stellar surface "
                                    "gravity log(g[cm/s^2]): [1.0, 5.0]",
                                    "espda.espda_synth_spec", 4.438);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "synth-logg");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /* Parameter synth_feh */
    param = cpl_parameter_new_value("espda.espda_synth_spec.synth_feh",
                                    CPL_TYPE_DOUBLE, "Stellar metallicity "
                                    "[Fe/H]: [-5.0, +0.5]",
                                    "espda.espda_synth_spec", 0.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "synth-feh");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

   /* Parameters of espda_convol_synth */
   /* Parameter synth_vins */
    param = cpl_parameter_new_value(
                                 "espda.espda_synth_spec.synth_vins",
                                    CPL_TYPE_DOUBLE, "Instrumental "
			            "broadening [km/s]: [0., 15.]", 
                                    "espda.espda_synth_spec", 2.1414);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "synth-vins");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);
   /* Parameter synth_vrot */
    param = cpl_parameter_new_value(
                                 "espda.espda_synth_spec.synth_vrot",
                                    CPL_TYPE_DOUBLE, "Stellar rotational "
			            "velocity [km/s]: [0., 300.]",
                                    "espda.espda_synth_spec", 1.6);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "synth-vrot");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

   /* Parameter synth_vmac */
    param = cpl_parameter_new_value(
                                 "espda.espda_synth_spec.synth_vmac",
				 CPL_TYPE_DOUBLE, "Stellar Macroturbulent "
                                  "velocity [km/s]: [0., 10.]", 
                                  "espda.espda_synth_spec", 3.2);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "synth-vmac");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_synth_spec_exec(cpl_plugin *plugin)
{
    int             recipe_status;

    cpl_recipe     *recipe;                                                   
    cpl_errorstate  initial_errorstate = cpl_errorstate_get();              
                                                                           
    /* Return immediately if an error code is already set */               
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                          
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",    
                      cpl_func, __LINE__, cpl_error_get_where());          
        return (int)cpl_error_get_code();                                  
    }                                                                      
                                                                           
    if (plugin == NULL) {                                                  
        cpl_msg_error(cpl_func, "Null plugin");                            
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    /* Verify plugin type */                                               
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {           
        cpl_msg_error(cpl_func, "Plugin is not a recipe");                 
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);                  
    }                                                                      
                                                                           
    /* Get the recipe */                                                   
    recipe = (cpl_recipe *)plugin;                                         
                                                                           
    /* Verify parameter and frame lists */                                 
    if (recipe->parameters == NULL) {                                      
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
    if (recipe->frames == NULL) {                                          
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");     
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    /* Invoke the recipe */                                                
    recipe_status = espda_synth_spec(recipe->frames, recipe->parameters);
                                                                           
    /* Ensure DFS-compliance of the products */                            
    if (cpl_dfs_update_product_header(recipe->frames)) {                   
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();                         
    }                                                                      
                                                                           
    if (!cpl_errorstate_is_equal(initial_errorstate)) {                    
        /* Dump the error history since recipe execution start.            
           At this point the recipe cannot recover from the error */       
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);          
    }                                                                      
                                                                           
    return recipe_status;                                                  
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_synth_spec_destroy(cpl_plugin *plugin)
{
    cpl_recipe * recipe;                                          
                                                                  
    if (plugin == NULL) {                                         
        cpl_msg_error(cpl_func, "Null plugin");                   
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);            
    }                                                             
                                                                  
    /* Verify plugin type */                                      
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {  
        cpl_msg_error(cpl_func, "Plugin is not a recipe");        
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);         
    }                                                             
                                                                  
    /* Get the recipe */                                          
    recipe = (cpl_recipe *)plugin;                                
                                                                  
    cpl_parameterlist_delete(recipe->parameters);             
                                                                  
    return 0;                                                    
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist    the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_synth_spec(cpl_frameset            *frameset,
			   const cpl_parameterlist *paramlist)
{
    cpl_propertylist *proplist   = NULL;

    cpl_table        *nmerg_tab   = NULL;
    cpl_table        *synth_tab   = NULL;

    cpl_msg_info(cpl_func, "*** Starting Synth Spec ***");

/*
 * Getting parameters from the list into variables 
 */

    espda_param_double("teff",         teff,    
        cpl_parameterlist_get_first_const (paramlist), 3500.,  7000.);

    espda_param_double("logg",         logg,    
        cpl_parameterlist_get_next_const (paramlist), 1.0,  5.0);

    espda_param_double("feh",         feh,    
        cpl_parameterlist_get_next_const (paramlist), -5.0,  +0.5);

    espda_param_double("vins",         vins,
        cpl_parameterlist_get_next_const (paramlist), 0.,  15.0);

    espda_param_double("vrot",         vrot,    
        cpl_parameterlist_get_next_const (paramlist), 0.,  300.0);

    espda_param_double("vmac",         vmac,
        cpl_parameterlist_get_next_const (paramlist), 0.,  10.0);

/* 
 * Spectra are imported from frame
 */

    cpl_msg_info(cpl_func, "*** Load input FITS frames into tables ***");
    cpl_ensure_code(espda_frame_import(frameset, NMERG_SPEC, &nmerg_tab) 
                    == CPL_ERROR_NONE, cpl_error_get_code());
    if (nmerg_tab == NULL) {
        return (int)cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND, 
               "No input merged spectrum frame! Terminating.");
    }
    cpl_msg_info(cpl_func, "*** 1D NORMALIZED MERGED SPEC - NMERG_SPEC table LOADED***");

/* 
 * Spectra are imported from frame
 */

   cpl_msg_info(cpl_func, "*** Load input FITS frames into tables ***");
   cpl_ensure_code(espda_frame_import(frameset, SYNTH_SPEC, &synth_tab) 
                    == CPL_ERROR_NONE, cpl_error_get_code());
    if (synth_tab == NULL) {
        return (int)cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND, 
               "No input synth spectra frame! Terminating.");
    }
    cpl_msg_info(cpl_func, "*** SYNTH GRID - SYNTH_SPEC table LOADED***");
   
    //    cpl_table_dump (synth_tab, 1, 10, NULL);

/*
 * DFS groups are set
 */

    cpl_msg_info(cpl_func, "*** Set groups of FITS frames ***");
    cpl_ensure_code(espda_dfs_groups(frameset) == CPL_ERROR_NONE,
                    cpl_error_get_code());

/*
 * Selecting 8 spectra within the Synthetic Grid to interpolate
 */ 

    printf("Defining the grid of synthetic spectra \n");

// Synthetic Teff values
const int nt = 8 ;
    cpl_array *tef_in  = cpl_array_new (nt, CPL_TYPE_DOUBLE);
    for (int j = 0; j < (int)nt; j++) {
      double ct = 3500. + 500. * (double)j;
      cpl_array_set_double (tef_in, j, ct);
    }
    cpl_array *mtef_in  = cpl_array_new (nt, CPL_TYPE_STRING);
      const char * mt1 = "T35";
      int j = 0;
      cpl_array_set_string (mtef_in, j, mt1);
      const char * mt2 = "T40";
      j++;
      cpl_array_set_string (mtef_in, j, mt2);
      const char * mt3 = "T45";
      j++;
      cpl_array_set_string (mtef_in, j, mt3);
      const char * mt4 = "T50";
      j++;
      cpl_array_set_string (mtef_in, j, mt4);
      const char * mt5 = "T55";
      j++;
      cpl_array_set_string (mtef_in, j, mt5);
      const char * mt6 = "T60";
      j++;
      cpl_array_set_string (mtef_in, j, mt6);
      const char * mt7 = "T65";
      j++;
      cpl_array_set_string (mtef_in, j, mt7);
      const char * mt8 = "T70";
      j++;
      cpl_array_set_string (mtef_in, j, mt8);


// Synthetic Logg values
    const int ng = 5 ;
    cpl_array *gra_in  = cpl_array_new (ng, CPL_TYPE_DOUBLE);
    for (int j = 0; j < (int)ng; j++) {
      double cg = 1.0 + 1.0 * (double)j;
      cpl_array_set_double (gra_in, j, cg);
    }
    cpl_array *mgra_in  = cpl_array_new (ng, CPL_TYPE_STRING);
      const char * mg1 = "G10";
      int h = 0;
      cpl_array_set_string (mgra_in, h, mg1);
      const char * mg2 = "G20";
      h++;
      cpl_array_set_string (mgra_in, h, mg2);
      const char * mg3 = "G30";
      h++;
      cpl_array_set_string (mgra_in, h, mg3);
      const char * mg4 = "G40";
      h++;
      cpl_array_set_string (mgra_in, h, mg4);
      const char * mg5 = "G50";
      h++;
      cpl_array_set_string (mgra_in, h, mg5);


// Synthetic [Fe/H] values
    const int nm = 8 ;
    //    printf(" Fe/H values: %i %f \n", nm, feh);
    cpl_array *feh_in  = cpl_array_new (nm, CPL_TYPE_DOUBLE);
    for (int j = 0; j < (int)nm-3; j++) {
      double cm = -5.0 + 1. * (double)j;
      //      printf(" Fe/H values: %f \n", cm);
      cpl_array_set_double (feh_in, j, cm);
    }
    j=5;
    double cm = -0.5;
    cpl_array_set_double (feh_in, j, cm);
    j=j+1;
    cm = 0.0;
    cpl_array_set_double (feh_in, j, cm);
    j=j+1;
    cm = 0.5;
    cpl_array_set_double (feh_in, j, cm);

      cpl_array *mfeh_in  = cpl_array_new (nm, CPL_TYPE_STRING);
      const char * mm1 = "M50";
      int k = 0;
      cpl_array_set_string (mfeh_in, k, mm1);
      const char * mm2 = "M40";
      k++;
      cpl_array_set_string (mfeh_in, k, mm2);
      const char * mm3 = "M30";
      k++;
      cpl_array_set_string (mfeh_in, k, mm3);
      const char * mm4 = "M20";
      k++;
      cpl_array_set_string (mfeh_in, k, mm4);
      const char * mm5 = "M10";
      k++;
      cpl_array_set_string (mfeh_in, k, mm5);
      const char * mm6 = "M05";
      k++;
      cpl_array_set_string (mfeh_in, k, mm6);
      const char * mm7 = "P00";
      k++;
      cpl_array_set_string (mfeh_in, k, mm7);
      const char * mm8 = "P05";
      k++;
      cpl_array_set_string (mfeh_in, k, mm8);

// Identifying the 8 model synthetic spectra
    
    printf("Identifying the grid of synthetic spectra \n");

    //Teff 
    int it1 = 0 ;
    for (int j = 0; j < (int)nt; j++) {
      double cmt = cpl_array_get_double (tef_in, j, NULL);
      if (teff > cmt) {
	it1 = j ; 
      }    
    }
    int it2 = it1 +1 ;

    double t1 = cpl_array_get_double (tef_in, it1, NULL);
    double t2 = cpl_array_get_double (tef_in, it2, NULL);
    printf(" Teff values: %f %f \n", t1, t2);

    const char * mtf1 = cpl_array_get_string (mtef_in, it1);
    const char * mtf2 = cpl_array_get_string (mtef_in, it2);
    printf(" Teff values: %s %s \n", mtf1, mtf2);

    //logg 
    int ig1 = 0 ;
    for (int j = 0; j < (int)ng; j++) {
      double cmg = cpl_array_get_double (gra_in, j, NULL);
      if (logg > cmg) {
	ig1 = j ; 
      }    
    }
    int ig2 = ig1 +1 ;
    double g1 = cpl_array_get_double (gra_in, ig1, NULL);
    double g2 = cpl_array_get_double (gra_in, ig2, NULL);
    printf(" Logg values: %f %f \n", g1, g2);

    const char * mgf1 = cpl_array_get_string (mgra_in, ig1);
    const char * mgf2 = cpl_array_get_string (mgra_in, ig2);
    printf(" Logg values: %s %s \n", mgf1, mgf2);

    //[Fe/H]
    int im1 = 0 ;
    for (int j = 0; j < (int)nm; j++) {
      double cmm = cpl_array_get_double (feh_in, j, NULL);
      //      printf(" Fe/H values: %f \n", cmm);
      if (feh > cmm) {
	im1 = j ; 
	//      printf(" Fe/H index: %i \n", im1);
      }    
    }
    int im2 = im1 +1 ;
    double m1 = cpl_array_get_double (feh_in, im1, NULL);
    double m2 = cpl_array_get_double (feh_in, im2, NULL);
    printf(" [Fe/H] values: %f %f \n", m1, m2);

    printf(" [Fe/H] indexes: %i %i \n", im1, im2);
    const char * mmf1 = cpl_array_get_string (mfeh_in, im1);
    const char * mmf2 = cpl_array_get_string (mfeh_in, im2);
    printf(" [Fe/H] values: %s %s \n", mmf1, mmf2);

    //    const char * mnam1 = mtf1 + mgf1 + mmf1;
    //    const char * mnam2 = mtf1 + mgf1 + mmf2;
    //    const char * mnam3 = mtf1 + mgf2 + mmf1;
    //    const char * mnam4 = mtf1 + mgf2 + mmf2;
    //    const char * mnam5 = mtf2 + mgf1 + mmf1;
    //    const char * mnam6 = mtf2 + mgf1 + mmf2;
    //    const char * mnam7 = mtf2 + mgf2 + mmf1;
    //    const char * mnam8 = mtf2 + mgf2 + mmf2;
    //printf(" Len values: %d %d %d \n",strlen(mtf1), strlen(mgf1), strlen(mmf1) );
    char * mna1 = malloc( (strlen(mtf1) + strlen(mgf1) + strlen(mmf1)+1) * sizeof(char) );
    strcpy (mna1,mtf1);
    strcat (mna1,mgf1);
    strcat (mna1,mmf1);
    char * mna2 = malloc( (strlen(mtf1) + strlen(mgf1) + strlen(mmf2)+1) * sizeof(char) );
    strcpy (mna2,mtf1);
    strcat (mna2,mgf1);
    strcat (mna2,mmf2);
    char * mna3 = malloc( (strlen(mtf1) + strlen(mgf2) + strlen(mmf1)+1) * sizeof(char) );
    strcpy (mna3,mtf1);
    strcat (mna3,mgf2);
    strcat (mna3,mmf1);
    char * mna4 = malloc( (strlen(mtf1) + strlen(mgf2) + strlen(mmf2)+1) * sizeof(char) );
    strcpy (mna4,mtf1);
    strcat (mna4,mgf2);
    strcat (mna4,mmf2);
    char * mna5 = malloc( (strlen(mtf2) + strlen(mgf1) + strlen(mmf1)+1) * sizeof(char) );
    strcpy (mna5,mtf2);
    strcat (mna5,mgf1);
    strcat (mna5,mmf1);
    char * mna6 = malloc( (strlen(mtf2) + strlen(mgf1) + strlen(mmf2)+1) * sizeof(char) );
    strcpy (mna6,mtf2);
    strcat (mna6,mgf1);
    strcat (mna6,mmf2);
    char * mna7 = malloc( (strlen(mtf2) + strlen(mgf2) + strlen(mmf1)+1) * sizeof(char) );
    strcpy (mna7,mtf2);
    strcat (mna7,mgf2);
    strcat (mna7,mmf1);
    char * mna8 = malloc( (strlen(mtf2) + strlen(mgf2) + strlen(mmf2)+1) * sizeof(char) );
    strcpy (mna8,mtf2);
    strcat (mna8,mgf2);
    strcat (mna8,mmf2);
        
    const char * mnam1 = mna1;
    const char * mnam2 = mna2;
    const char * mnam3 = mna3;
    const char * mnam4 = mna4;
    const char * mnam5 = mna5;
    const char * mnam6 = mna6;
    const char * mnam7 = mna7;
    const char * mnam8 = mna8;

    printf("Selected Synthetic model fluxes: \n");
    printf("Model 1: %s \n", mnam1);
    printf("Model 2: %s \n", mnam2);
    printf("Model 3: %s \n", mnam3);
    printf("Model 4: %s \n", mnam4);
    printf("Model 5: %s \n", mnam5);
    printf("Model 6: %s \n", mnam6);
    printf("Model 7: %s \n", mnam7);
    printf("Model 8: %s \n", mnam8);

// extracting wave of the normalized merged spectrum

    const double *wobsp = cpl_table_get_data_double_const (nmerg_tab, "WAVEL");
    cpl_array *wobs_arr = cpl_array_new(cpl_table_get_nrow(nmerg_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (wobs_arr, wobsp);

// extracting wave of the normalized merged spectrum

    const double *fobsp = cpl_table_get_data_double_const (nmerg_tab, "FLUX");
    cpl_array *fobs_arr = cpl_array_new(cpl_table_get_nrow(nmerg_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (fobs_arr, fobsp);

// extracting wave of the grid of Synthetic spectra

    const double *wsynp = cpl_table_get_data_double_const (synth_tab, "WSYNTH");
    cpl_array *wsyn_arr = cpl_array_new(cpl_table_get_nrow(synth_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (wsyn_arr, wsynp);

// extracting flux of the grid of Synthetic spectra

    printf("Selecting the flux arrays from the grid of synthetic spectra \n");

    const double *fsynp1 = cpl_table_get_data_double_const (synth_tab, mnam1);
    cpl_array *fsyn1_arr = cpl_array_new(cpl_table_get_nrow(synth_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (fsyn1_arr, fsynp1);
   
    const double *fsynp2 = cpl_table_get_data_double_const (synth_tab, mnam2);
    cpl_array *fsyn2_arr = cpl_array_new(cpl_table_get_nrow(synth_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (fsyn2_arr, fsynp2);

    const double *fsynp3 = cpl_table_get_data_double_const (synth_tab, mnam3);
    cpl_array *fsyn3_arr = cpl_array_new(cpl_table_get_nrow(synth_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (fsyn3_arr, fsynp3);

    const double *fsynp4 = cpl_table_get_data_double_const (synth_tab, mnam4);
    cpl_array *fsyn4_arr = cpl_array_new(cpl_table_get_nrow(synth_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (fsyn4_arr, fsynp4);

    const double *fsynp5 = cpl_table_get_data_double_const (synth_tab, mnam5);
    cpl_array *fsyn5_arr = cpl_array_new(cpl_table_get_nrow(synth_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (fsyn5_arr, fsynp5);

    const double *fsynp6 = cpl_table_get_data_double_const (synth_tab, mnam6);
    cpl_array *fsyn6_arr = cpl_array_new(cpl_table_get_nrow(synth_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (fsyn6_arr, fsynp6);

    const double *fsynp7 = cpl_table_get_data_double_const (synth_tab, mnam7);
    cpl_array *fsyn7_arr = cpl_array_new(cpl_table_get_nrow(synth_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (fsyn7_arr, fsynp7);

    const double *fsynp8 = cpl_table_get_data_double_const (synth_tab, mnam8);
    cpl_array *fsyn8_arr = cpl_array_new(cpl_table_get_nrow(synth_tab), CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double (fsyn8_arr, fsynp8);

// Calling the function to interpolate the synthetic spectra

    cpl_array *synth_arr = cpl_array_new(cpl_table_get_nrow(nmerg_tab), CPL_TYPE_DOUBLE);

    printf("Interpolating the synthetic spectrum for the stellar parameters: \n");
    printf("Teff = %f, logg = %f, [Fe/H]= %f \n", teff, logg, feh);
          
    cpl_ensure_code(espda_interp_synth(wobs_arr, wsyn_arr, fsyn1_arr, fsyn2_arr, 
                        fsyn3_arr, fsyn4_arr, fsyn5_arr, fsyn6_arr, fsyn7_arr,
                        fsyn8_arr, synth_arr, t1, t2, g1, g2, m1, m2, teff, logg, 
	                    feh) == CPL_ERROR_NONE, cpl_error_get_code());

 // Calling the function to convolve the synthetic spectrum with
 // different broadening profiles

     cpl_array *sconv_arr = cpl_array_new(cpl_table_get_nrow(nmerg_tab), CPL_TYPE_DOUBLE);

     printf("Convolving the synthetic spectrum for the broadening parameters: \n");
     printf("Instrumental broadening = %f \n", vins);
     printf("Rotational velocity vsini= %f \n", vrot);
     printf("Macroturbulent velocity = %f \n", vmac);

     cpl_ensure_code(espda_convol_comp(wobs_arr, synth_arr, sconv_arr,
                            vins, vrot, vmac,
							teff) == CPL_ERROR_NONE, cpl_error_get_code());

 // Create array of the residuals Observed - Synthetic spectrum SYNTH

    cpl_array *sreso_arr = cpl_array_duplicate(fobs_arr);
    cpl_array_subtract (sreso_arr, synth_arr);

    cpl_array *sresy_arr = cpl_array_duplicate(fobs_arr);
    cpl_array_subtract (sresy_arr, sconv_arr);

 // Copy the output contiuum data into a table DOUBLE column labelled as "SYNTH"

    //printf("Copying the output flux array into the SYNTH table column \n");

// We duplicate the table
    cpl_table *smerg_tab = cpl_table_duplicate (nmerg_tab);

// We create the Table columns "SYNTH" and "SCONV"
    cpl_table_new_column(smerg_tab, "SYNTO", CPL_TYPE_DOUBLE);
    cpl_table_new_column(smerg_tab, "SYNTH", CPL_TYPE_DOUBLE);
    cpl_table_new_column(smerg_tab, "RESIO", CPL_TYPE_DOUBLE);
    cpl_table_new_column(smerg_tab, "RESID", CPL_TYPE_DOUBLE);

// We copy the synthetic spectrum into columns "SYNTH" and "SCONV"
    const double *synthf = cpl_array_get_data_double_const (synth_arr);
    cpl_table_copy_data_double (smerg_tab, "SYNTO", synthf);
    const double *sresof = cpl_array_get_data_double_const (sreso_arr);
    cpl_table_copy_data_double (smerg_tab, "RESIO", sresof);

    const double *sconvf = cpl_array_get_data_double_const (sconv_arr);
    cpl_table_copy_data_double (smerg_tab, "SYNTH", sconvf);
    const double *sresyf = cpl_array_get_data_double_const (sresy_arr);
    cpl_table_copy_data_double (smerg_tab, "RESID", sresyf);

    /* We delete the remaining arrays to release and clean the memory */
    cpl_array_delete(fsyn1_arr);
    cpl_array_delete(fsyn2_arr);
    cpl_array_delete(fsyn3_arr);
    cpl_array_delete(fsyn4_arr);
    cpl_array_delete(fsyn5_arr);
    cpl_array_delete(fsyn6_arr);
    cpl_array_delete(fsyn7_arr);
    cpl_array_delete(fsyn8_arr);

    cpl_array_delete(wsyn_arr);
    cpl_array_delete(fobs_arr);
    cpl_array_delete(wobs_arr);

    cpl_array_delete(tef_in);
    cpl_array_delete(gra_in);
    cpl_array_delete(feh_in);
    cpl_array_delete(mtef_in);
    cpl_array_delete(mgra_in);
    cpl_array_delete(mfeh_in);

    cpl_array_delete(synth_arr);
    cpl_array_delete(sconv_arr);
    cpl_array_delete(sreso_arr);
    cpl_array_delete(sresy_arr);

/*
 * Products are exported into frames. 
 */ 

    cpl_msg_info(cpl_func, "*** Export products into frames ***");

    cpl_table_and_selected_double (smerg_tab, "WAVEL", CPL_LESS_THAN, 680.);
    cpl_table *smerger_tab = cpl_table_extract_selected (smerg_tab);
    cpl_table_unselect_all (smerg_tab);

    proplist = cpl_propertylist_new();
	cpl_propertylist_append_double(proplist, "ESO DAS QC SYNTH Teff", teff);
    cpl_propertylist_append_double(proplist, "ESO DAS QC SYNTH Logg", logg);
	cpl_propertylist_append_double(proplist, "ESO DAS QC SYNTH FeH", feh);
	cpl_propertylist_append_double(proplist, "ESO DAS QC SYNTH Vins", vins);
	cpl_propertylist_append_double(proplist, "ESO DAS QC SYNTH Vmac", vmac);
	cpl_propertylist_append_double(proplist, "ESO DAS QC SYNTH Vrot", vrot);
    cpl_propertylist_append_string(proplist, "ESO DAS QC SYNTH MODEL1", mnam1);
    cpl_propertylist_append_string(proplist, "ESO DAS QC SYNTH MODEL2", mnam2);
    cpl_propertylist_append_string(proplist, "ESO DAS QC SYNTH MODEL3", mnam3);
    cpl_propertylist_append_string(proplist, "ESO DAS QC SYNTH MODEL4", mnam4);
    cpl_propertylist_append_string(proplist, "ESO DAS QC SYNTH MODEL5", mnam5);
    cpl_propertylist_append_string(proplist, "ESO DAS QC SYNTH MODEL6", mnam6);
    cpl_propertylist_append_string(proplist, "ESO DAS QC SYNTH MODEL7", mnam7);
    cpl_propertylist_append_string(proplist, "ESO DAS QC SYNTH MODEL8", mnam8);
    double mmflx= cpl_table_get_column_mean(smerger_tab, "FLUX") ;
    cpl_propertylist_append_double(proplist, "ESO DAS QC SYNTH MEAN MFLUX W680", mmflx);
    double msflx= cpl_table_get_column_mean(smerger_tab, "SYNTH") ;
    cpl_propertylist_append_double(proplist, "ESO DAS QC SYNTH MEAN FSYNTH W680", msflx);
    double mrflx= cpl_table_get_column_mean(smerger_tab, "RESID") ;
    cpl_propertylist_append_double(proplist, "ESO DAS QC SYNTH MEAN RESID W680", mrflx);
    double mrstd= cpl_table_get_column_stdev(smerger_tab, "RESID") ;
    cpl_propertylist_append_double(proplist, "ESO DAS QC SYNTH STDEV RESID W680", mrstd);
    if (smerg_tab != NULL) {
     cpl_ensure_code(espda_frame_export(frameset, paramlist, smerg_tab, 
                        proplist, "espda_synth_spec", SMERG_SPEC, 
                        "SMERG_SPEC.fits")  == CPL_ERROR_NONE,
                        cpl_error_get_code());
    }
    else {
     cpl_msg_warning(cpl_func, "No synthetic spectrum interpolated! No frame exported.");
    }

/*    
*Finishing the recipe
*/

    /* We delete the remaining variables to release and clean the memory */
    cpl_free(mna1);
    cpl_free(mna2);
    cpl_free(mna3);
    cpl_free(mna4);
    cpl_free(mna5);
    cpl_free(mna6);
    cpl_free(mna7);
    cpl_free(mna8);

// We delete all the property lists to release and clean the memory
//    cpl_propertylist_delete(proplist);

// We delete all the tables to release and clean the memory
    cpl_table_delete(smerger_tab);
    cpl_table_delete(nmerg_tab);
    cpl_table_delete(synth_tab);

    cpl_msg_info(cpl_func, "*** Successful Recipe Execution!! ***");

    return (int)cpl_error_get_code();

}

/**@}*/
/*--------------------------------------------------------------------------*/
